# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import json
import logging
import warnings
from typing import TYPE_CHECKING, ClassVar, Optional, Union

from typing_extensions import Literal

try:
    from pydantic.v1 import Field, root_validator, validator
except (ImportError, AttributeError):
    from pydantic import Field, validator, root_validator  # type: ignore[no-redef, assignment]

from onetl._util.java import try_import_java_class
from onetl._util.scala import get_default_scala_version
from onetl._util.spark import get_spark_version
from onetl._util.version import Version
from onetl.exception import MISSING_JVM_CLASS_MSG
from onetl.file.format.file_format import ReadWriteFileFormat
from onetl.hooks import slot, support_hooks

if TYPE_CHECKING:
    from pyspark.sql import Column, DataFrameReader, DataFrameWriter, SparkSession


PROHIBITED_OPTIONS = frozenset(
    (
        # These options should be passed in Spark session config, not file format options
        "spark.*",
    ),
)

PARSE_COLUMN_UNSUPPORTED_OPTIONS = {
    "compression",
}

SERIALIZE_COLUMN_UNSUPPORTED_OPTIONS = {
    "compression",
    "mode",
    "positionalFieldMatching",
    "enableStableIdentifiersForUnionType",
}

log = logging.getLogger(__name__)


@support_hooks
class Avro(ReadWriteFileFormat):
    """
    Avro file format. |support_hooks|

    Based on `Spark Avro <https://spark.apache.org/docs/latest/sql-data-sources-avro.html>`_ file format.

    Supports reading/writing files with ``.avro`` extension.

    .. dropdown:: Version compatibility

        * Spark versions: 2.4.x - 3.5.x
        * Java versions: 8 - 20

        See documentation from link above.

    .. versionadded:: 0.9.0

    Examples
    --------

    .. note ::

        You can pass any option mentioned in
        `official documentation <https://spark.apache.org/docs/latest/sql-data-sources-avro.html>`_.
        **Option names should be in** ``camelCase``!

        The set of supported options depends on Spark version.

    .. tabs::

        .. code-tab:: py Reading files

            from pyspark.sql import SparkSession
            from onetl.file.format import Avro

            # Create Spark session with Avro package loaded
            maven_packages = Avro.get_packages(spark_version="3.5.6")
            spark = (
                SparkSession.builder.appName("spark-app-name")
                .config("spark.jars.packages", ",".join(maven_packages))
                .getOrCreate()
            )

            schema = {
                "type": "record",
                "name": "Person",
                "fields": [
                    {"name": "name", "type": "string"},
                    {"name": "age", "type": "int"},
                ],
            }
            avro = Avro(avroSchema=schema)  # or avroSchemaUrl=...

        .. code-tab:: py Writing files

            # Create Spark session with Avro package loaded
            spark = ...

            from onetl.file.format import Avro

            schema = {
                "type": "record",
                "name": "Person",
                "fields": [
                    {"name": "name", "type": "string"},
                    {"name": "age", "type": "int"},
                ],
            }
            avro = Avro(
                avroSchema=schema,  # or avroSchemaUrl=...
                compression="snappy",
            )

    """

    name: ClassVar[str] = "avro"

    schema_dict: Optional[dict] = Field(default=None, alias="avroSchema")
    """
    Avro schema in JSON format representation.

    .. code:: python

        avro = Avro(
            avroSchema={
                "type": "record",
                "name": "Person",
                "fields": [
                    {"name": "name", "type": "string"},
                    {"name": "age", "type": "int"},
                ],
            },
        )

    If set, all records should match this schema.

    .. warning::

        Mutually exclusive with :obj:`~schema_url`.
    """

    schema_url: Optional[str] = Field(default=None, alias="avroSchemaUrl")
    """
    URL to Avro schema in JSON format. Usually points to Schema Registry, like:

    .. code:: python

        schema_registry = "http://some.schema.registry.domain"
        name = "MyAwesomeSchema"
        version = "latest"

        schema_url = f"{schema_registry}/subjects/{name}/versions/{version}/schema"
        avro = Avro(avroSchemaUrl=schema_url)

    If set, schema is fetched before any records are parsed, so all records should match this schema.

    .. warning::

        Mutually exclusive with :obj:`~schema_dict`.
    """

    recordName: Optional[str] = None
    """
    Record name in written Avro schema.
    Default is ``topLevelRecord``.

    .. note::

        Used only for writing files and by :obj:`~serialize_column`.
    """

    recordNamespace: Optional[str] = None
    """
    Record namespace in written Avro schema. Default is not set.

    .. note::

        Used only for writing files and by :obj:`~serialize_column`.
    """

    compression: Union[str, Literal["uncompressed", "snappy", "deflate", "bzip2", "xz", "zstandard"], None] = None
    """
    Compression codec.
    By default, Spark config value ``spark.sql.avro.compression.codec `` (``snappy``) is used.

    .. note::

        Used only for writing files. Ignored by :obj:`~serialize_column`.
    """

    mode: Optional[Literal["PERMISSIVE", "FAILFAST"]] = None
    """
    How to handle parsing errors:
      * ``PERMISSIVE`` - set field value as ``null``.
      * ``FAILFAST`` - throw an error immediately.

    Default is ``FAILFAST``.

    .. note::

        Used only by :obj:`~parse_column` method.
    """

    datetimeRebaseMode: Optional[Literal["CORRECTED", "LEGACY", "EXCEPTION"]] = None
    """
    While converting dates/timestamps from Julian to Proleptic Gregorian calendar, handle value ambiguity:
      * ``EXCEPTION`` - fail if ancient dates/timestamps are ambiguous between the two calendars.
      * ``CORRECTED`` - load dates/timestamps without as-is.
      * ``LEGACY`` - rebase ancient dates/timestamps from the Julian to Proleptic Gregorian calendar.

    By default, Spark config value ``spark.sql.avro.datetimeRebaseModeInRead`` (``CORRECTED``) is used.

    .. note::

        Used only for reading files and by :obj:`~parse_column`.
    """

    positionalFieldMatching: Optional[bool] = None
    """
    If ``True``, match Avro schema field and DataFrame column by position.
    If ``False``, match by name.

    Default is ``False``.
    """

    enableStableIdentifiersForUnionType: Optional[bool] = None
    """
    Avro schema may contain union types, which are not supported by Spark.
    Different variants of union are split to separated DataFrame columns with respective type.

    If option value is ``True``, DataFrame column names are based on Avro variant names, e.g. ``member_int``, ``member_string``.
    If ``False``, DataFrame column names are generated using field position, e.g. ``member0``, ``member1``.

    Default is ``False``.

    .. note::

        Used only for reading files and by :obj:`~parse_column`.
    """

    class Config:
        known_options = frozenset()
        prohibited_options = PROHIBITED_OPTIONS
        extra = "allow"

    @slot
    @classmethod
    def get_packages(
        cls,
        spark_version: str,
        scala_version: str | None = None,
    ) -> list[str]:
        """
        Get package names to be downloaded by Spark. |support_hooks|

        See `Maven package index <https://mvnrepository.com/artifact/org.apache.spark/spark-avro>`_
        for all available packages.

        .. versionadded:: 0.9.0

        Parameters
        ----------
        spark_version : str
            Spark version in format ``major.minor.patch``.

        scala_version : str, optional
            Scala version in format ``major.minor``.

            If ``None``, ``spark_version`` is used to determine Scala version.

        Examples
        --------

        .. code:: python

            from onetl.file.format import Avro

            Avro.get_packages(spark_version="3.5.6")
            Avro.get_packages(spark_version="3.5.6", scala_version="2.12")

        """

        spark_ver = Version(spark_version).min_digits(3)
        scala_ver = Version(scala_version).min_digits(2) if scala_version else get_default_scala_version(spark_ver)
        return [f"org.apache.spark:spark-avro_{scala_ver.format('{0}.{1}')}:{spark_ver.format('{0}.{1}.{2}')}"]

    @slot
    def check_if_supported(self, spark: SparkSession) -> None:
        java_class = "org.apache.spark.sql.avro.AvroFileFormat"

        try:
            try_import_java_class(spark, java_class)
        except Exception as e:
            spark_version = get_spark_version(spark).format("{0}.{1}.{2}")
            msg = MISSING_JVM_CLASS_MSG.format(
                java_class=java_class,
                package_source=self.__class__.__name__,
                args=f"spark_version='{spark_version}'",
            )
            raise ValueError(msg) from e

    @slot
    def apply_to_reader(self, reader: DataFrameReader) -> DataFrameReader:
        options = self.dict(by_alias=True, exclude_none=True, exclude={"schema"})
        if self.schema_dict:
            options["avroSchema"] = json.dumps(self.schema_dict)
        return reader.format(self.name).options(**options)

    @slot
    def apply_to_writer(self, writer: DataFrameWriter) -> DataFrameWriter:
        options = self.dict(by_alias=True, exclude_none=True, exclude={"schema"})
        if self.schema_dict:
            options["avroSchema"] = json.dumps(self.schema_dict)
        return writer.format(self.name).options(**options)

    def parse_column(self, column: str | Column) -> Column:
        """
        Parses an Avro binary column into a structured Spark SQL column using Spark's
        `from_avro <https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.avro.functions.from_avro.html>`_ function,
        based on the schema provided within the class.

        .. note::

            Can be used only with Spark 3.x+

        .. warning::

            If ``schema_url`` is provided, ``requests`` library is used to fetch the schema from the URL.
            It should be installed manually, like this:

            .. code:: bash

                pip install requests

        .. versionadded:: 0.11.0

        Parameters
        ----------
        column : str | Column
            The name of the column or the column object containing Avro bytes to deserialize.
            Schema should match the provided Avro schema.

        Returns
        -------
        Column with deserialized data. Schema is matching the provided Avro schema. Column name is the same as input column.

        Raises
        ------
        ValueError
            If the Spark version is less than 3.x or if neither ``avroSchema`` nor ``avroSchemaUrl`` are defined.
        ImportError
            If ``schema_url`` is used and the ``requests`` library is not installed.

        Examples
        --------

        >>> from pyspark.sql.functions import decode
        >>> from onetl.file.format import Avro
        >>> df.show()
        +----+----------------------+----------+---------+------+-----------------------+-------------+
        |key |value                 |topic     |partition|offset|timestamp              |timestampType|
        +----+----------------------+----------+---------+------+-----------------------+-------------+
        |[31]|[0A 41 6C 69 63 65 28]|topicAvro |0        |0     |2024-04-24 13:02:25.911|0            |
        |[32]|[06 42 6F 62 32]      |topicAvro |0        |1     |2024-04-24 13:02:25.922|0            |
        +----+----------------------+----------+---------+------+-----------------------+-------------+
        >>> df.printSchema()
        root
        |-- key: binary (nullable = true)
        |-- value: binary (nullable = true)
        |-- topic: string (nullable = true)
        |-- partition: integer (nullable = true)
        |-- offset: integer (nullable = true)
        |-- timestamp: timestamp (nullable = true)
        |-- timestampType: integer (nullable = true)
        >>> avro = Avro(
        ...     avroSchema={  # or avroSchemaUrl=...
        ...         "type": "record",
        ...         "name": "Person",
        ...         "fields": [
        ...             {"name": "name", "type": "string"},
        ...             {"name": "age", "type": "int"},
        ...         ],
        ...     }
        ... )
        >>> parsed_df = df.select(decode("key", "UTF-8").alias("key"), avro.parse_column("value"))
        >>> parsed_df.show(truncate=False)
        +---+-----------+
        |key|value      |
        +---+-----------+
        |1  |{Alice, 20}|
        |2  |{Bob, 25}  |
        +---+-----------+
        >>> parsed_df.printSchema()
        root
        |-- key: string (nullable = true)
        |-- value: struct (nullable = true)
        |    |-- name: string (nullable = true)
        |    |-- age: integer (nullable = true)
        """
        from pyspark.sql import Column, SparkSession  # noqa: WPS442
        from pyspark.sql.functions import col

        spark = SparkSession._instantiatedSession  # noqa: WPS437
        self.check_if_supported(spark)
        self._check_unsupported_parse_options()

        from pyspark.sql.avro.functions import from_avro

        if isinstance(column, Column):
            column_name = column._jc.toString()  # noqa: WPS437
        else:
            column_name, column = column, col(column).cast("binary")

        schema = self._get_schema_json()
        if not schema:
            raise ValueError("Avro.parse_column can be used only with defined `avroSchema` or `avroSchemaUrl`")

        return from_avro(column, schema).alias(column_name)

    def serialize_column(self, column: str | Column) -> Column:
        """
        Serializes a structured Spark SQL column into an Avro binary column using Spark's
        `to_avro <https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.avro.functions.to_avro.html#pyspark.sql.avro.functions.to_avro>`_ function.

        .. note::

            Can be used only with Spark 3.x+

        .. warning::

            If ``schema_url`` is provided, ``requests`` library is used to fetch the schema from the URL. It should be installed manually, like this:

            .. code:: bash

                pip install requests

        .. versionadded:: 0.11.0

        Parameters
        ----------
        column : str | Column
            The name of the column or the column object containing the data to serialize to Avro format.

        Returns
        -------
        Column with binary Avro data. Column name is the same as input column.

        Raises
        ------
        ValueError
            If the Spark version is less than 3.x.
        ImportError
            If ``schema_url`` is used and the ``requests`` library is not installed.

        Examples
        --------

        >>> from pyspark.sql.functions import decode
        >>> from onetl.file.format import Avro
        >>> df.show()
        +---+-----------+
        |key|value      |
        +---+-----------+
        |1  |{Alice, 20}|
        |2  |  {Bob, 25}|
        +---+-----------+
        >>> df.printSchema()
        root
        |-- key: string (nullable = true)
        |-- value: struct (nullable = true)
        |    |-- name: string (nullable = true)
        |    |-- age: integer (nullable = true)
        >>> # serializing data into Avro format
        >>> avro = Avro(
        ...     avroSchema={  # or avroSchemaUrl=...
        ...         "type": "record",
        ...         "name": "Person",
        ...         "fields": [
        ...             {"name": "name", "type": "string"},
        ...             {"name": "age", "type": "int"},
        ...         ],
        ...     }
        ... )
        >>> serialized_df = df.select("key", avro.serialize_column("value"))
        >>> serialized_df.show(truncate=False)
        +---+----------------------+
        |key|value                 |
        +---+----------------------+
        |  1|[0A 41 6C 69 63 65 28]|
        |  2|[06 42 6F 62 32]      |
        +---+----------------------+
        >>> serialized_df.printSchema()
        root
        |-- key: string (nullable = true)
        |-- value: binary (nullable = true)
        """
        from pyspark.sql import Column, SparkSession  # noqa: WPS442
        from pyspark.sql.functions import col

        spark = SparkSession._instantiatedSession  # noqa: WPS437
        self.check_if_supported(spark)
        self._check_unsupported_serialization_options()

        from pyspark.sql.avro.functions import to_avro

        if isinstance(column, Column):
            column_name = column._jc.toString()  # noqa:  WPS437
        else:
            column_name, column = column, col(column)

        schema = self._get_schema_json()
        return to_avro(column, schema).alias(column_name)

    def __repr__(self):
        options_dict = self.dict(by_alias=True, exclude_none=True)
        options_kwargs = ", ".join(f"{k}={v!r}" for k, v in options_dict.items())
        return f"{self.__class__.__name__}({options_kwargs})"

    def _check_unsupported_parse_options(self):
        current_options = self.dict(by_alias=True, exclude_none=True)
        unsupported_options = current_options.keys() & PARSE_COLUMN_UNSUPPORTED_OPTIONS
        if unsupported_options:
            warnings.warn(
                f"Options `{sorted(unsupported_options)}` are set but not supported by `Avro.parse_column`.",
                UserWarning,
                stacklevel=2,
            )

    def _check_unsupported_serialization_options(self):
        current_options = self.dict(by_alias=True, exclude_none=True)
        unsupported_options = current_options.keys() & SERIALIZE_COLUMN_UNSUPPORTED_OPTIONS
        if unsupported_options:
            warnings.warn(
                f"Options `{sorted(unsupported_options)}` are set but not supported by `Avro.serialize_column`.",
                UserWarning,
                stacklevel=2,
            )

    @validator("schema_dict", pre=True)
    def _parse_schema_from_json(cls, value):
        if isinstance(value, (str, bytes)):
            return json.loads(value)
        return value

    @root_validator(pre=True)
    def _check_schema(cls, values):
        schema_dict = values.get("schema_dict")
        schema_url = values.get("schema_url")
        if schema_dict and schema_url:
            raise ValueError("Parameters `avroSchema` and `avroSchemaUrl` are mutually exclusive.")
        return values

    def _get_schema_json(self) -> str:
        if self.schema_dict:
            return json.dumps(self.schema_dict)
        elif self.schema_url:
            try:
                import requests

                response = requests.get(self.schema_url)  # noqa: S113
                return response.text
            except ImportError as e:
                raise ImportError(
                    "The 'requests' library is required to use 'schema_url' but is not installed. "
                    "Install it with 'pip install requests' or avoid using 'schema_url'.",
                ) from e
        else:
            return ""



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from onetl.connection.db_connection.iceberg.connection import Iceberg
from onetl.connection.db_connection.iceberg.dialect import IcebergDialect
from onetl.connection.db_connection.iceberg.extra import IcebergExtra
from onetl.connection.db_connection.iceberg.options import (
    IcebergReadOptions,
    IcebergWriteOptions,
)



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from typing_extensions import Literal

try:
    from pydantic.v1 import Field
except (ImportError, AttributeError):
    from pydantic import Field  # type: ignore[no-redef, assignment]

from onetl.file.format.file_format import ReadWriteFileFormat
from onetl.hooks import slot, support_hooks

if TYPE_CHECKING:
    from pyspark.sql import SparkSession


@support_hooks
class JSONLine(ReadWriteFileFormat):
    """
    JSONLine file format (each line of file contains a JSON object). |support_hooks|

    Based on `Spark JSON <https://spark.apache.org/docs/latest/sql-data-sources-json.html>`_ file format.

    Supports reading/writing files with ``.json`` extension with content like:

    .. code-block:: json
        :caption: example.json

        {"key": "value1"}
        {"key": "value2"}

    .. versionadded:: 0.9.0

    Examples
    --------

    .. note ::

        You can pass any option mentioned in
        `official documentation <https://spark.apache.org/docs/latest/sql-data-sources-json.html>`_.
        **Option names should be in** ``camelCase``!

        The set of supported options depends on Spark version.

    .. tabs::

        .. code-tab:: py Reading files

            from onetl.file.format import JSONLine

            jsonline = JSONLine(encoding="UTF-8", mode="PERMISSIVE")

        .. tab:: Writing files

            .. warning::

                Written files have extension ``.json``, not ``.jsonl`` or ``.jsonline``.

            .. code:: python

                from onetl.file.format import JSONLine

                jsonline = JSONLine(encoding="UTF-8", compression="gzip")
    """

    name: ClassVar[str] = "json"

    multiLine: Literal[False] = False

    encoding: Optional[str] = None
    """
    Encoding of the JSONLine files.
    Default ``UTF-8``.
    """

    lineSep: Optional[str] = None
    """
    Character used to separate lines in the JSONLine files.

    Defaults:
      * Try to detect for reading (``\\r\\n``, ``\\r``, ``\\n``)
      * ``\\n`` for writing.
    """

    compression: Union[str, Literal["none", "bzip2", "gzip", "lz4", "snappy", "deflate"], None] = None
    """
    Compression codec of the JSONLine file.
    Default ``none``.

    .. note::

        Used only for writing files.
    """

    ignoreNullFields: Optional[bool] = None
    """
    If ``True`` and field value is ``null``, don't add field into resulting object
    Default is value of ``spark.sql.jsonGenerator.ignoreNullFields`` (``True``).

    .. note::

        Used only for writing files.
    """

    allowComments: Optional[bool] = None
    """
    If ``True``, add support for C/C++/Java style comments (``//``, ``/* */``).
    Default ``False``, meaning that JSONLine files should not contain comments.

    .. note::

        Used only for reading files.
    """

    allowUnquotedFieldNames: Optional[bool] = None
    """
    If ``True``, allow JSON object field names without quotes (JavaScript-style).
    Default ``False``.

    .. note::

        Used only for reading files.
    """

    allowSingleQuotes: Optional[bool] = None
    """
    If ``True``, allow JSON object field names to be wrapped with single quotes (``'``).
    Default ``True``.

    .. note::

        Used only for reading files.
    """

    allowNumericLeadingZeros: Optional[bool] = None
    """
    If ``True``, allow leading zeros in numbers (e.g. ``00012``).
    Default ``False``.

    .. note::

        Used only for reading files.
    """

    allowNonNumericNumbers: Optional[bool] = None
    """
    If ``True``, allow numbers to contain non-numeric characters, like:
      * scientific notation (e.g. ``12e10``).
      * positive infinity floating point value (``Infinity``, ``+Infinity``, ``+INF``).
      * negative infinity floating point value (``-Infinity``, ``-INF``).
      * Not-a-Number floating point value (``NaN``).

    Default ``True``.

    .. note::

        Used only for reading files.
    """

    allowBackslashEscapingAnyCharacter: Optional[bool] = None
    """
    If ``True``, prefix ``\\`` can escape any character.
    Default ``False``.

    .. note::

        Used only for reading files.
    """

    allowUnquotedControlChars: Optional[bool] = None
    """
    If ``True``, allow unquoted control characters (ASCII values 0-31) in strings without escaping them with ``\\``.
    Default ``False``.

    .. note::

        Used only for reading files.
    """

    mode: Optional[Literal["PERMISSIVE", "DROPMALFORMED", "FAILFAST"]] = None
    """
    How to handle parsing errors:
      * ``PERMISSIVE`` - set field value as ``null``, move raw data to :obj:`~columnNameOfCorruptRecord` column.
      * ``DROPMALFORMED`` - skip the malformed row.
      * ``FAILFAST`` - throw an error immediately.

    Default is ``PERMISSIVE``.

    .. note::

        Used only for reading files.
    """

    columnNameOfCorruptRecord: Optional[str] = Field(default=None, min_length=1)
    """
    Name of column to put corrupt records in.
    Default is ``_corrupt_record``.

    .. warning::

        If DataFrame schema is provided, this column should be added to schema explicitly:

        .. code:: python

            from onetl.connection import SparkLocalFS
            from onetl.file import FileDFReader
            from onetl.file.format import JSONLine

            from pyspark.sql.types import StructType, StructField, TimestampType, StringType

            spark = ...

            schema = StructType(
                [
                    StructField("my_field", TimestampType()),
                    StructField("_corrupt_record", StringType()),  # <-- important
                ]
            )

            jsonline = JSONLine(mode="PERMISSIVE", columnNameOfCorruptRecord="_corrupt_record")

            reader = FileDFReader(
                connection=connection,
                format=jsonline,
                df_schema=schema,  # < ---
            )
            df = reader.run(["/some/file.jsonl"])

    .. note::

        Used only for reading files.
    """

    samplingRatio: Optional[float] = Field(default=None, ge=0, le=1)
    """
    While inferring schema, read the specified fraction of file rows.
    Default ``1``.

    .. note::

        Used only for reading files.
    """

    primitivesAsString: Optional[bool] = None
    """
    If ``True``, infer all primitive types (string, integer, float, boolean) as strings.
    Default ``False``.

    .. note::

        Used only for reading files.
    """

    prefersDecimal: Optional[bool] = None
    """
    If ``True``, infer all floating-point values as ``Decimal``.
    Default ``False``.

    .. note::

        Used only for reading files.
    """

    dropFieldIfAllNull: Optional[bool] = None
    """
    If ``True`` and inferred column is always null or empty array, exclude if from DataFrame schema.
    Default ``False``.

    .. note::

        Used only for reading files.
    """

    dateFormat: Optional[str] = Field(default=None, min_length=1)
    """
    String format for ``DateType()`` representation.
    Default is ``yyyy-MM-dd``.
    """

    timestampFormat: Optional[str] = Field(default=None, min_length=1)
    """
    String format for `TimestampType()`` representation.
    Default is ``yyyy-MM-dd'T'HH:mm:ss[.SSS][XXX]``.
    """

    timestampNTZFormat: Optional[str] = Field(default=None, min_length=1)
    """
    String format for `TimestampNTZType()`` representation.
    Default is ``yyyy-MM-dd'T'HH:mm:ss[.SSS]``.

    .. note::

        Added in Spark 3.2.0
    """

    timezone: Optional[str] = Field(default=None, min_length=1, alias="timeZone")
    """
    Allows to override timezone used for parsing or serializing date and timestamp values.
    By default, ``spark.sql.session.timeZone`` is used.
    """

    locale: Optional[str] = Field(default=None, min_length=1)
    """
    Locale name used to parse dates and timestamps.
    Default is ``en-US``.

    ..  note::

        Used only for reading files.
    """

    class Config:
        known_options = frozenset()
        extra = "allow"

    @slot
    def check_if_supported(self, spark: SparkSession) -> None:
        # always available
        pass

    def __repr__(self):
        options_dict = self.dict(by_alias=True, exclude_none=True, exclude={"multiLine"})
        options_dict = dict(sorted(options_dict.items()))
        options_kwargs = ", ".join(f"{k}={v!r}" for k, v in options_dict.items())
        return f"{self.__class__.__name__}({options_kwargs})"



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0

from onetl.connection.db_connection.jdbc_connection.options import (
    JDBCReadOptions,
    JDBCSQLOptions,
    JDBCWriteOptions,
)
from onetl.connection.db_connection.jdbc_mixin.options import (
    JDBCExecuteOptions,
    JDBCFetchOptions,
)


class PostgresReadOptions(JDBCReadOptions):
    __doc__ = JDBCReadOptions.__doc__.replace("SomeDB", "Postgres")  # type: ignore[assignment, union-attr]


class PostgresWriteOptions(JDBCWriteOptions):
    __doc__ = JDBCWriteOptions.__doc__.replace("SomeDB", "Postgres")  # type: ignore[assignment, union-attr]


class PostgresSQLOptions(JDBCSQLOptions):
    __doc__ = JDBCSQLOptions.__doc__.replace("SomeDB", "Postgres")  # type: ignore[assignment, union-attr]


class PostgresFetchOptions(JDBCFetchOptions):
    __doc__ = JDBCFetchOptions.__doc__.replace("SomeDB", "Postgres")  # type: ignore[assignment, union-attr]


class PostgresExecuteOptions(JDBCExecuteOptions):
    __doc__ = JDBCExecuteOptions.__doc__.replace("SomeDB", "Postgres")  # type: ignore[assignment, union-attr]



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from enum import Enum

from onetl._util.alias import avoid_alias

try:
    from pydantic.v1 import Field
except (ImportError, AttributeError):
    from pydantic import Field  # type: ignore[no-redef, assignment]

from onetl.impl import GenericOptions

PROHIBITED_OPTIONS = frozenset(
    (
        # These options should be passed in Spark session config
        "spark.*",
    ),
)


class IcebergTableExistBehavior(str, Enum):
    APPEND = "append"
    IGNORE = "ignore"
    ERROR = "error"
    REPLACE_ENTIRE_TABLE = "replace_entire_table"
    REPLACE_OVERLAPPING_PARTITIONS = "replace_overlapping_partitions"

    def __str__(self):
        return str(self.value)


class IcebergReadOptions(GenericOptions):
    """Iceberg source reading options."""

    class Config:
        extra = "allow"
        known_options: frozenset = frozenset()
        prohibited_options = PROHIBITED_OPTIONS


class IcebergWriteOptions(GenericOptions):
    """Iceberg source writing options."""

    class Config:
        extra = "allow"
        known_options: frozenset = frozenset()
        prohibited_options = PROHIBITED_OPTIONS

    if_exists: IcebergTableExistBehavior = Field(  # type: ignore[literal-required]
        default=IcebergTableExistBehavior.APPEND,
        alias=avoid_alias("mode"),
    )
    """Behavior of writing data into existing table.

    Possible values:
        * ``append`` (default)
            Appends data into existing table, or create table if it does not exist.

            Same as Spark's ``df.writeTo(table).using("iceberg").append()``.

            .. dropdown:: Behavior in details

                * Table does not exist
                    Table is created.

                * Table exists and not partitioned
                    Data is appended to a table. Table DDL (including partition spec) is unchanged.

                * Table exists and partitioned
                    If a partition is present only in dataframe
                        Partition is created.
                    If a partition is present in both dataframe and table
                        Data is appended to existing partition.

                    .. warning::

                        This mode does not check whether table already contains
                        rows from dataframe, so duplicated rows can be created.

                        To implement deduplication, write data to staging table first,
                        and then perform some deduplication logic using :obj:`~sql`.

                * Table exists and partitioned, but some partitions are present only in table, not dataframe
                    Existing partitions are left intact.

        * ``replace_overlapping_partitions``
            Overwrites data in the existing partitions, or create table if it does not exist.

            Same as Spark's ``df.writeTo(table).using("iceberg").overwritePartitions()``

            .. DANGER::

                This mode does make sense **ONLY** if the table is partitioned.
                **IF NOT, YOU'LL LOSE YOUR DATA!**

            .. dropdown:: Behavior in details

                * Table does not exist
                    Table is created.

                * Table exists and not partitioned
                    Data is **overwritten in all the table**. Table DDL (including partition spec) is unchanged.

                * Table exists and partitioned
                    If a partition is present only in dataframe
                        Partition is created.
                    If a partition is present in both dataframe and table
                        Existing partition **replaced** with data from dataframe.
                    If a partition is present only in table, not dataframe
                        Existing partition is left intact.

        * ``replace_entire_table``
            **Recreates table** (via ``DROP + CREATE``), **deleting all existing data**.
            **All existing partitions are dropped.**

            Same as Spark's ``df.writeTo(table).createOrReplace()``

            .. warning::

                Table is recreated
                **instead of using original table options**. Be careful

        * ``ignore``
            Ignores the write operation if the table already exists.

            .. dropdown:: Behavior in details

                * Table does not exist
                    Table is created.

                * Table exists
                    If the table exists, **no further action is taken**.

        * ``error``
            Raises an error if the table already exists.

            .. dropdown:: Behavior in details

                * Table does not exist
                    Table is created.

                * Table exists
                    If the table exists, **raises an error**.
    """



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, ClassVar, Optional, Union

from typing_extensions import Literal

try:
    from pydantic.v1 import Field
except (ImportError, AttributeError):
    from pydantic import Field  # type: ignore[no-redef, assignment]

from onetl._util.alias import avoid_alias
from onetl._util.spark import stringify
from onetl.file.format.file_format import ReadWriteFileFormat
from onetl.hooks import slot, support_hooks

if TYPE_CHECKING:
    from pyspark.sql import Column, SparkSession
    from pyspark.sql.types import StructType

PARSE_COLUMN_UNSUPPORTED_OPTIONS = {
    "header",
    "compression",
    "encoding",
    "inferSchema",
    "samplingRatio",
    "enforceSchema",
    "preferDate",
    "multiLine",
}


@support_hooks
class CSV(ReadWriteFileFormat):
    """
    CSV file format. |support_hooks|

    Based on `Spark CSV <https://spark.apache.org/docs/latest/sql-data-sources-csv.html>`_ file format.

    Supports reading/writing files with ``.csv`` extension with content like:

    .. code-block:: csv
        :caption: example.csv

        "some","value"
        "another","value"

    .. versionadded:: 0.9.0

    Examples
    --------

    .. note ::

        You can pass any option mentioned in
        `official documentation <https://spark.apache.org/docs/latest/sql-data-sources-csv.html>`_.
        **Option names should be in** ``camelCase``!

        The set of supported options depends on Spark version.

    .. tabs::

        .. code-tab:: py Reading files

            from onetl.file.format import CSV

            csv = CSV(header=True, inferSchema=True, mode="PERMISSIVE")

        .. code-tab:: py Writing files

            from onetl.file.format import CSV

            csv = CSV(header=True, compression="gzip")

    """

    name: ClassVar[str] = "csv"

    delimiter: str = Field(default=",", alias=avoid_alias("sep"))  # type: ignore[literal-required]
    """
    Character used to separate fields in CSV row.
    """

    header: Optional[bool] = None
    """
    If ``True``, the first row of the file is considered a header.
    Default ``False``.
    """

    quote: str = Field(default='"', max_length=1)
    """
    Character used to quote field values within CSV field.

    Empty string is considered as ``\\u0000`` (``NUL``) character.
    """

    quoteAll: Optional[bool] = None
    """
    If ``True``, all fields are quoted:

    .. code:: csv

        "some","field with \\"quote","123",""

    If ``False``, only quote fields containing :obj:`~quote` symbols.

    .. code:: csv

        any,"field with \\"quote",123,

    Default ``False``.

    .. note::

        Used only for writing files.
    """

    escape: str = Field(default="\\", max_length=1)
    """
    Character used to escape quotes within CSV field.

    Empty string is considered as ``\\u0000`` (``NUL``) character.
    """

    lineSep: Optional[str] = Field(default=None, min_length=1, max_length=2)
    """
    Character used to separate lines in the CSV file.

    Defaults:
      * Try to detect for reading (``\\r\\n``, ``\\r``, ``\\n``)
      * ``\\n`` for writing

    .. note::

        Used only for reading and writing files.
        Ignored by :obj:`~parse_column` method, as it expects that each DataFrame row will contain exactly one CSV line.
    """

    encoding: Optional[str] = Field(default=None, min_length=1)
    """
    Encoding of the CSV file.
    Default ``UTF-8``.

    .. note::

        Used only for reading and writing files. Ignored by :obj:`~parse_column` method.
    """

    compression: Union[str, Literal["none", "bzip2", "gzip", "lz4", "snappy", "deflate"], None] = None
    """
    Compression codec of the CSV file.
    Default ``none``.

    .. note::

        Used only for writing files. Ignored by :obj:`~parse_column` method.
    """

    inferSchema: Optional[bool] = None
    """
    If ``True``, try to infer the input schema by reading a sample of the file (see :obj:`~samplingRatio`).
    Default ``False`` which means that all parsed columns will be ``StringType()``.

    .. note::

        Used only for reading files, and only if user haven't provider explicit DataFrame schema.
        Ignored by :obj:`~parse_column` function.
    """

    samplingRatio: Optional[float] = Field(default=None, ge=0, le=1)
    """
    For ``inferSchema=True``, read the specified fraction of rows to infer the schema.
    Default ``1``.

    .. note::

        Used only for reading files. Ignored by :obj:`~parse_column` function.
    """

    comment: Optional[str] = Field(default=None, max_length=1)
    """
    If set, all lines starting with specified character (e.g. ``#``) are considered a comment, and skipped.
    Default is not set, meaning that CSV lines should not contain comments.

    .. note::

        Used only for reading files and :obj:`~parse_column` method.
    """

    enforceSchema: Optional[bool] = None
    """
    If ``True``, inferred or user-provided schema has higher priority than CSV file headers.
    This means that all input files should have the same structure.

    If ``False``, CSV headers are used as a primary source of information about column names and their position.

    Default ``True``.

    .. note::

        Used only for reading files. Ignored by :obj:`~parse_column` function.
    """

    escapeQuotes: Optional[bool] = None
    """
    If ``True``, escape quotes within CSV field.

    .. code:: csv

        any,field with \\"quote,123,

    If ``False``, wrap fields containing :obj:`~quote` symbols with quotes.

    .. code:: csv

        any,"field with ""quote ",123,

    Default ``True``.

    .. note::

        Used only for writing files.
    """

    unescapedQuoteHandling: Union[
        None,
        Literal[
            "STOP_AT_CLOSING_QUOTE",
            "BACK_TO_DELIMITER",
            "STOP_AT_DELIMITER",
            "SKIP_VALUE",
            "RAISE_ERROR",
        ],
    ] = None
    """
    Define how to handle unescaped quotes within CSV field.

    * ``STOP_AT_CLOSING_QUOTE`` - collect all characters until closing quote.
    * ``BACK_TO_DELIMITER`` - collect all characters until delimiter or line end.
    * ``STOP_AT_DELIMITER`` - collect all characters until delimiter or line end.
       If quotes are not closed, this may produce incorrect results (e.g. including delimiter inside field value).
    * ``SKIP_VALUE`` - skip field and consider it as :obj:`~nullValue`.
    * ``RAISE_ERROR`` - raise error immediately.

    Default ``STOP_AT_DELIMITER``.

    .. note::

        Used only for reading files and :obj:`~parse_column` method.
    """

    ignoreLeadingWhiteSpace: Optional[bool] = None
    """
    If ``True``, trim leading whitespaces in field value.

    Defaults:
      * ``True`` for reading.
      * ``False`` for writing.
    """

    ignoreTrailingWhiteSpace: Optional[bool] = None
    """
    If ``True``, trim trailing whitespaces in field value.

    Defaults:
      * ``True`` for reading.
      * ``False`` for writing.
    """

    emptyValue: Optional[str] = None
    """
    Value used for empty string fields.

    Defaults:
      * empty string for reading.
      * ``""`` for writing.
    """

    nullValue: Optional[str] = None
    """
    If set, this value will be converted to ``null``.
    Default is empty string.
    """

    nanValue: Optional[str] = Field(default=None)
    """
    If set, this string will be considered as Not-A-Number (NaN) value for ``FloatType()`` and ``DoubleType()``.
    Default is ``NaN``.

    .. note::

        Used only for reading files and :obj:`~parse_column` method.
    """

    positiveInf: Optional[str] = Field(default=None, min_length=1)
    """
    If set, this string will be considered as positive infinity value for ``FloatType()`` and ``DoubleType()``.
    Default is ``Inf``.

    .. note::

        Used only for reading files and :obj:`~parse_column` method.
    """

    negativeInf: Optional[str] = Field(default=None, min_length=1)
    """
    If set, this string will be considered as negative infinity value for ``FloatType()`` and ``DoubleType()``.
    Default is ``-Inf``.

    .. note::

        Used only for reading files and :obj:`~parse_column` method.
    """

    preferDate: Optional[bool] = None
    """
    If ``True`` and ``inferSchema=True`` and column does match :obj:`~dateFormat`, consider it as ``DateType()``.
    For columns matching both :obj:`~timestampFormat` and :obj:`~dateFormat`, consider it as ``TimestampType()``.

    If ``False``, date and timestamp columns will be considered as ``StringType()``.

    Default ``True``.

    .. note::

        Used only for reading files. Ignored by :obj:`~parse_column` function.
    """

    dateFormat: Optional[str] = Field(default=None, min_length=1)
    """
    String format for ``DateType()`` representation.
    Default is ``yyyy-MM-dd``.
    """

    timestampFormat: Optional[str] = Field(default=None, min_length=1)
    """
    String format for `TimestampType()`` representation.
    Default is ``yyyy-MM-dd'T'HH:mm:ss[.SSS][XXX]``.
    """

    timestampNTZFormat: Optional[str] = Field(default=None, min_length=1)
    """
    String format for `TimestampNTZType()`` representation.
    Default is ``yyyy-MM-dd'T'HH:mm:ss[.SSS]``.

    .. note::

        Added in Spark 3.2.0
    """

    locale: Optional[str] = Field(default=None, min_length=1)
    """
    Locale name used to parse dates and timestamps.
    Default is ``en-US``

    ..  note::

        Used only for reading files and :obj:`~parse_column` method.
    """

    maxCharsPerColumn: Optional[int] = None
    """
    Maximum number of characters to read per column.
    Default is ``-1``, which means no limit.

    .. note::

        Used only for reading files and :obj:`~parse_column` method.
    """

    mode: Optional[Literal["PERMISSIVE", "DROPMALFORMED", "FAILFAST"]] = None
    """
    How to handle parsing errors:
      * ``PERMISSIVE`` - set field value as ``null``, move raw data to :obj:`~columnNameOfCorruptRecord` column.
      * ``DROPMALFORMED`` - skip the malformed row.
      * ``FAILFAST`` - throw an error immediately.

    Default is ``PERMISSIVE``.

    .. note::

        Used only for reading files and :obj:`~parse_column` method.
    """

    columnNameOfCorruptRecord: Optional[str] = Field(default=None, min_length=1)
    """
    Name of column to put corrupt records in.
    Default is ``_corrupt_record``.

    .. warning::

        If DataFrame schema is provided, this column should be added to schema explicitly:

        .. code:: python

            from onetl.connection import SparkLocalFS
            from onetl.file import FileDFReader
            from onetl.file.format import CSV

            from pyspark.sql.types import StructType, StructField, TimestampType, StringType

            spark = ...

            schema = StructType(
                [
                    StructField("my_field", TimestampType()),
                    StructField("_corrupt_record", StringType()),  # <-- important
                ]
            )

            csv = CSV(mode="PERMISSIVE", columnNameOfCorruptRecord="_corrupt_record")

            reader = FileDFReader(
                connection=connection,
                format=csv,
                df_schema=schema,  # < ---
            )
            df = reader.run(["/some/file.csv"])

    .. note::

        Used only for reading files and :obj:`~parse_column` method.
    """

    multiLine: Optional[bool] = None
    """
    If ``True``, fields may contain line separators.
    If ``False``, the input is expected to have one record per file.

    Default is ``True``.

    .. note::

        Used only for reading files.
        Ignored by :obj:`~parse_column` method, as it expects that each DataFrame row will contain exactly one CSV line.
    """

    charToEscapeQuoteEscaping: Optional[str] = Field(default=None, max_length=1)
    """
    If CSV field value contains :obj:`~escape` character, it should be escaped as well.
    For example, if ``escape="\\"``, when line:

    .. code:: csv

        "some \\" quoted value",other
        "some \\\\ backslashed value",another

    will be parsed as:

    .. code:: python

        [
            ('some " quoted value', "other"),
            ("some \\ backslashed value", "another"),
        ]

    And vice-versa, for writing CSV rows to file.

    Default is same as :obj:`~escape`.
    """

    class Config:
        known_options = frozenset()
        extra = "allow"

    @slot
    @classmethod
    def check_if_supported(cls, spark: SparkSession) -> None:
        # always available
        pass

    def parse_column(self, column: str | Column, schema: StructType) -> Column:
        """
        Parses a CSV string column to a structured Spark SQL column using Spark's
        `from_csv <https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.functions.from_csv.html>`_ function, based on the provided schema.

        .. note::

            Can be used only with Spark 3.x+

        .. versionadded:: 0.11.0

        Parameters
        ----------
        column : str | Column
            The name of the column or the column object containing CSV strings/bytes to parse.

        schema : StructType
            The schema to apply when parsing the CSV data. This defines the structure of the output DataFrame column.

        Returns
        -------
        Column with deserialized data, with the same structure as the provided schema. Column name is the same as input column.

        Examples
        --------

        >>> from pyspark.sql.types import StructType, StructField, IntegerType, StringType
        >>> from onetl.file.format import CSV
        >>> df.show()
        +--+--------+
        |id|value   |
        +--+--------+
        |1 |Alice;20|
        |2 |Bob;25  |
        +--+--------+
        >>> df.printSchema()
        root
        |-- id: integer (nullable = true)
        |-- value: string (nullable = true)
        >>> csv = CSV(delimiter=";")
        >>> csv_schema = StructType(
        ...     [
        ...         StructField("name", StringType(), nullable=True),
        ...         StructField("age", IntegerType(), nullable=True),
        ...     ],
        ... )
        >>> parsed_df = df.select("id", csv.parse_column("value", csv_schema))
        >>> parsed_df.show()
        +--+-----------+
        |id|value      |
        +--+-----------+
        |1 |{Alice, 20}|
        |2 |  {Bob, 25}|
        +--+-----------+
        >>> parsed_df.printSchema()
        root
        |-- id: integer (nullable = true)
        |-- value: struct (nullable = true)
        |    |-- name: string (nullable = true)
        |    |-- age: integer (nullable = true)
        """

        from pyspark.sql import Column, SparkSession  # noqa: WPS442

        spark = SparkSession._instantiatedSession  # noqa: WPS437
        self.check_if_supported(spark)
        self._check_unsupported_serialization_options()

        from pyspark.sql.functions import col, from_csv

        if isinstance(column, Column):
            column_name = column._jc.toString()  # noqa: WPS437
        else:
            column_name, column = column, col(column).cast("string")

        schema_string = schema.simpleString()
        options = stringify(self.dict(by_alias=True, exclude_none=True))
        return from_csv(column, schema_string, options).alias(column_name)

    def serialize_column(self, column: str | Column) -> Column:
        """
        Serializes a structured Spark SQL column into a CSV string column using Spark's
        `to_csv <https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.functions.to_csv.html>`_ function.

        .. note::

            Can be used only with Spark 3.x+

        .. versionadded:: 0.11.0

        Parameters
        ----------
        column : str | Column
            The name of the column or the Column object containing the data to serialize to CSV.

        Returns
        -------
        Column with string CSV data. Column name is the same as input column.

        Examples
        --------

        >>> from pyspark.sql.functions import decode
        >>> from onetl.file.format import CSV
        >>> df.show()
        +--+-----------+
        |id|value      |
        +--+-----------+
        |1 |{Alice, 20}|
        |2 |  {Bob, 25}|
        +--+-----------+
        >>> df.printSchema()
        root
        |-- id: integer (nullable = true)
        |-- value: struct (nullable = true)
        |    |-- name: string (nullable = true)
        |    |-- age: integer (nullable = true)
        >>> # serializing data into CSV format
        >>> csv = CSV(delimiter=";")
        >>> serialized_df = df.select("id", csv.serialize_column("value"))
        >>> serialized_df.show(truncate=False)
        +--+--------+
        |id|value   |
        +--+--------+
        |1 |Alice;20|
        |2 |Bob;25  |
        +--+--------+
        >>> serialized_df.printSchema()
        root
        |-- id: integer (nullable = true)
        |-- value: string (nullable = true)
        """

        from pyspark.sql import Column, SparkSession  # noqa: WPS442

        spark = SparkSession._instantiatedSession  # noqa: WPS437
        self.check_if_supported(spark)
        self._check_unsupported_serialization_options()

        from pyspark.sql.functions import col, to_csv

        if isinstance(column, Column):
            column_name = column._jc.toString()  # noqa: WPS437
        else:
            column_name, column = column, col(column)

        options = stringify(self.dict(by_alias=True, exclude_none=True))
        return to_csv(column, options).alias(column_name)

    def _check_unsupported_serialization_options(self):
        current_options = self.dict(by_alias=True, exclude_none=True)
        unsupported_options = current_options.keys() & PARSE_COLUMN_UNSUPPORTED_OPTIONS
        if unsupported_options:
            warnings.warn(
                f"Options `{sorted(unsupported_options)}` are set but not supported in `CSV.parse_column` or `CSV.serialize_column`.",
                UserWarning,
                stacklevel=2,
            )

    def __repr__(self):
        options_dict = self.dict(by_alias=True, exclude_none=True)
        options_dict = dict(sorted(options_dict.items()))
        options_kwargs = ", ".join(f"{k}={v!r}" for k, v in options_dict.items())
        return f"{self.__class__.__name__}({options_kwargs})"



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Iterable

from onetl._util.java import try_import_java_class
from onetl._util.scala import get_default_scala_version
from onetl._util.version import Version
from onetl.connection.db_connection.iceberg.dialect import IcebergDialect
from onetl.connection.db_connection.iceberg.extra import IcebergExtra
from onetl.connection.db_connection.iceberg.options import (
    IcebergReadOptions,
    IcebergTableExistBehavior,
    IcebergWriteOptions,
)
from onetl.exception import MISSING_JVM_CLASS_MSG

try:
    from pydantic.v1 import validator
except (ImportError, AttributeError):
    from pydantic import validator  # type: ignore[no-redef, assignment]

from onetl._metrics.recorder import SparkMetricsRecorder
from onetl._util.spark import get_spark_version, override_job_description
from onetl._util.sql import clear_statement
from onetl.connection.db_connection.db_connection import DBConnection
from onetl.hooks import slot, support_hooks
from onetl.hwm import Window
from onetl.log import log_lines, log_with_indent

if TYPE_CHECKING:
    from pyspark.sql import DataFrame, SparkSession
    from pyspark.sql.types import StructType


log = logging.getLogger(__name__)


@support_hooks
class Iceberg(DBConnection):
    """Iceberg connection. |support_hooks|

    .. danger::

        This is an alpha version of connector, it's behavior may change in the future.

    .. seealso::

        Before using this connector please take into account :ref:`iceberg-prerequisites`

    .. versionadded:: 0.14.0

    Parameters
    ----------
    catalog_name : str
        Catalog name

    spark : :obj:`pyspark.sql.SparkSession`
        Spark session

    extra : dict | None, default: ``None``
        A dictionary of additional properties to be used when configuring Iceberg catalog.

        These are Iceberg-specific properties that control behavior of the catalog.
        See `Iceberg Spark configuration documentation <https://iceberg.apache.org/docs/latest/spark-configuration/>`_

        Pass properties **without catalog prefix**. For example:

        .. code:: python

            extra = {
                "type": "hadoop",
                "warehouse": "file:///path/to/warehouse",
            }

        This will be translated to:

        .. code:: ini

            spark.sql.catalog.my_catalog = 'org.apache.iceberg.spark.SparkCatalog'
            spark.sql.catalog.my_catalog.type = 'hadoop'
            spark.sql.catalog.my_catalog.warehouse = 'file:///path/to/warehouse'

    Examples
    --------

    **REST catalog + S3 warehouse**

    .. code:: python

        from onetl.connection import Iceberg
        from pyspark.sql import SparkSession

        maven_packages = [
            *Iceberg.get_packages(package_version="1.9.2", spark_version="3.5"),
            *SparkS3.get_packages(spark_version="3.5.6"),
        ]
        exclude_packages = SparkS3.get_exclude_packages()
        spark = (
            SparkSession.builder.appName("spark-app-name")
            .config("spark.jars.packages", ",".join(maven_packages))
            .config("spark.jars.excludes", ",".join(exclude_packages))
            .getOrCreate()
        )

        # Create connection
        iceberg = Iceberg(
            catalog_name="my_catalog",
            spark=spark,
            extra={
                "type": "rest",
                "uri": "http://localhost:8080",
                "warehouse": "s3a://bucket/",
                "hadoop.fs.s3a.endpoint": "http://localhost:9010",
                "hadoop.fs.s3a.access.key": "access_key",
                "hadoop.fs.s3a.secret.key": "secret_key",
                "hadoop.fs.s3a.path.style.access": "true",
            },
        )


    **Hadoop catalog + HDFS warehouse**

    .. code:: python

        from onetl.connection import Iceberg
        from pyspark.sql import SparkSession

        maven_packages = Iceberg.get_packages(package_version="1.9.2", spark_version="3.5")
        spark = (
            SparkSession.builder.appName("spark-app-name")
            .config("spark.jars.packages", ",".join(maven_packages))
            .getOrCreate()
        )

        # Create connection
        iceberg = Iceberg(
            catalog_name="my_catalog",
            spark=spark,
            extra={
                "type": "hadoop",
                "warehouse": "hdfs://namenode:8020/warehouse/path",
            },
        )
    """

    catalog_name: str
    extra: IcebergExtra = IcebergExtra()

    ReadOptions = IcebergReadOptions
    WriteOptions = IcebergWriteOptions

    Dialect = IcebergDialect

    @property
    def _check_query(self) -> str:
        return f"SHOW NAMESPACES IN {self.catalog_name}"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.spark.conf.set(
            f"spark.sql.catalog.{self.catalog_name}",
            "org.apache.iceberg.spark.SparkCatalog",
        )
        for k, v in self.extra.dict().items():
            self.spark.conf.set(f"spark.sql.catalog.{self.catalog_name}.{k}", v)

    @slot
    @classmethod
    def get_packages(
        cls,
        package_version: str,
        spark_version: str,
        scala_version: str | None = None,
    ) -> list[str]:
        """
        Get package names to be downloaded by Spark. |support_hooks|

        See `Maven package index <https://mvnrepository.com/artifact/org.apache.iceberg/iceberg-spark>`_
        for all available packages.

        Parameters
        ----------
        package_version : str
            Iceberg package version in format ``major.minor.patch``.

        spark_version : str
            Spark version in format ``major.minor``.

        scala_version : str, optional
            Scala version in format ``major.minor``.

            If ``None``, ``spark_version`` is used to determine Scala version.

        Returns
        -------
        list[str]
            List of Maven coordinates.

        Examples
        --------
        .. code:: python

            from onetl.connection import Iceberg

            # Note: Iceberg 1.9.2 requires Java 11+
            Iceberg.get_packages(package_version="1.9.2", spark_version="3.5")
        """

        version = Version(package_version).min_digits(3)
        spark_ver = Version(spark_version).min_digits(2)
        scala_ver = Version(scala_version).min_digits(2) if scala_version else get_default_scala_version(spark_ver)
        return [
            f"org.apache.iceberg:iceberg-spark-runtime-{spark_ver.format('{0}.{1}')}_{scala_ver.format('{0}.{1}')}:{version}",
        ]

    @property
    def instance_url(self) -> str:
        return f"{self.__class__.__name__.lower()}://{self.catalog_name}"

    def __str__(self):
        return f"{self.__class__.__name__}[{self.catalog_name}]"

    @validator("spark")
    def _check_java_class_imported(cls, spark: SparkSession) -> SparkSession:
        java_class = "org.apache.iceberg.spark.SparkSessionCatalog"

        try:
            try_import_java_class(spark, java_class)
        except Exception as e:
            spark_version = get_spark_version(spark).format("{0}.{1}.{2}")
            msg = MISSING_JVM_CLASS_MSG.format(
                java_class=java_class,
                package_source=cls.__class__.__name__,
                args=f"spark_version='{spark_version}'",
            )
            raise ValueError(msg) from e

        return spark

    @slot
    def check(self):
        log.info("|%s| Checking connection availability...", self.__class__.__name__)
        self._log_parameters()

        log.debug("|%s| Executing SQL query:", self.__class__.__name__)
        log_lines(log, self._check_query, level=logging.DEBUG)

        try:
            with override_job_description(self.spark, f"{self}.check()"):
                self._execute_sql(self._check_query).take(1)
            log.info("|%s| Connection is available.", self.__class__.__name__)
        except Exception as e:
            log.exception("|%s| Connection is unavailable", self.__class__.__name__)
            raise RuntimeError("Connection is unavailable") from e

        return self

    @slot
    def sql(
        self,
        query: str,
    ) -> DataFrame:
        """
        Lazily execute SELECT statement and return DataFrame. |support_hooks|

        Same as ``spark.sql(query)``.

        Parameters
        ----------
        query : str

            SQL query to be executed.

        Returns
        -------
        df : pyspark.sql.dataframe.DataFrame

            Spark dataframe
        """

        query = clear_statement(query)

        log.info("|%s| Executing SQL query:", self.__class__.__name__)
        log_lines(log, query)

        with SparkMetricsRecorder(self.spark) as recorder:
            try:
                with override_job_description(self.spark, f"{self}.sql()"):
                    df = self._execute_sql(query)
            except Exception:
                log.error("|%s| Query failed", self.__class__.__name__)

                metrics = recorder.metrics()
                if log.isEnabledFor(logging.DEBUG) and not metrics.is_empty:
                    # as SparkListener results are not guaranteed to be received in time,
                    # some metrics may be missing. To avoid confusion, log only in debug, and with a notice
                    log.info("|%s| Recorded metrics (some values may be missing!):", self.__class__.__name__)
                    log_lines(log, str(metrics), level=logging.DEBUG)
                raise

            log.info("|Spark| DataFrame successfully created from SQL statement")

            metrics = recorder.metrics()
            if log.isEnabledFor(logging.DEBUG) and not metrics.is_empty:
                # as SparkListener results are not guaranteed to be received in time,
                # some metrics may be missing. To avoid confusion, log only in debug, and with a notice
                log.info("|%s| Recorded metrics (some values may be missing!):", self.__class__.__name__)
                log_lines(log, str(metrics), level=logging.DEBUG)

        return df

    @slot
    def execute(
        self,
        statement: str,
    ) -> None:
        """
        Execute DDL or DML statement. |support_hooks|

        Parameters
        ----------
        statement : str

            Statement to be executed.
        """

        statement = clear_statement(statement)

        log.info("|%s| Executing statement:", self.__class__.__name__)
        log_lines(log, statement)

        with SparkMetricsRecorder(self.spark) as recorder:
            try:
                with override_job_description(self.spark, f"{self}.execute()"):
                    self._execute_sql(statement).collect()
            except Exception:
                log.error("|%s| Execution failed", self.__class__.__name__)
                metrics = recorder.metrics()
                if log.isEnabledFor(logging.DEBUG) and not metrics.is_empty:
                    # as SparkListener results are not guaranteed to be received in time,
                    # some metrics may be missing. To avoid confusion, log only in debug, and with a notice
                    log.info("|%s| Recorded metrics (some values may be missing!):", self.__class__.__name__)
                    log_lines(log, str(metrics), level=logging.DEBUG)
                raise

            log.info("|%s| Execution succeeded", self.__class__.__name__)

            metrics = recorder.metrics()
            if log.isEnabledFor(logging.DEBUG) and not metrics.is_empty:
                # as SparkListener results are not guaranteed to be received in time,
                # some metrics may be missing. To avoid confusion, log only in debug, and with a notice
                log.info("|%s| Recorded metrics (some values may be missing!):", self.__class__.__name__)
                log_lines(log, str(metrics), level=logging.DEBUG)

    @slot
    def write_df_to_target(
        self,
        df: DataFrame,
        target: str,
        options: IcebergWriteOptions | None = None,
    ) -> None:
        target = self._normalize_table_name(target)
        write_options = self.WriteOptions.parse(options)

        table_exists = self._target_exist(target)

        if not table_exists or write_options.if_exists == IcebergTableExistBehavior.REPLACE_ENTIRE_TABLE:
            self._save_as_table(df, target, options)
            return

        if write_options.if_exists == IcebergTableExistBehavior.ERROR:
            raise ValueError("Operation stopped due to Iceberg.WriteOptions(if_exists='error')")

        if write_options.if_exists == IcebergTableExistBehavior.IGNORE:
            log.info(
                "|%s| Skip writing to existing table because of Iceberg.WriteOptions(if_exists='ignore')",
                self.__class__.__name__,
            )
            return

        self._insert_into(df, target, options)

    @slot
    def read_source_as_df(
        self,
        source: str,
        columns: list[str] | None = None,
        hint: str | None = None,
        where: str | None = None,
        df_schema: StructType | None = None,
        window: Window | None = None,
        limit: int | None = None,
        options: IcebergReadOptions | None = None,
    ) -> DataFrame:
        query = self.dialect.get_sql_query(
            table=self._normalize_table_name(source),
            columns=columns,
            where=self.dialect.apply_window(where, window),
            hint=hint,
            limit=limit,
        )
        return self.sql(query)

    @slot
    def get_df_schema(
        self,
        source: str,
        columns: list[str] | None = None,
    ) -> StructType:
        source = self._normalize_table_name(source)
        log.info("|%s| Fetching schema of table %r ...", self.__class__.__name__, source)
        query = self.dialect.get_sql_query(source, columns=columns, where=0, compact=True)

        log.debug("|%s| Executing SQL query:", self.__class__.__name__)
        log_lines(log, query, level=logging.DEBUG)

        df = self._execute_sql(query)
        log.info("|%s| Schema fetched.", self.__class__.__name__)
        return df.schema

    @slot
    def get_min_max_values(
        self,
        source: str,
        window: Window,
        hint: Any | None = None,
        where: Any | None = None,
    ) -> tuple[Any, Any]:
        log.info("|%s| Getting min and max values for expression %r ...", self.__class__.__name__, window.expression)

        source = self._normalize_table_name(source)
        query = self.dialect.get_sql_query(
            table=source,
            columns=[
                self.dialect.aliased(
                    self.dialect.get_min_value(window.expression),
                    self.dialect.escape_column("min"),
                ),
                self.dialect.aliased(
                    self.dialect.get_max_value(window.expression),
                    self.dialect.escape_column("max"),
                ),
            ],
            where=self.dialect.apply_window(where, window),
            hint=hint,
        )

        log.info("|%s| Executing SQL query:", self.__class__.__name__)
        log_lines(log, query)

        df = self._execute_sql(query)
        row = df.collect()[0]
        min_value = row["min"]
        max_value = row["max"]

        log.info("|%s| Received values:", self.__class__.__name__)
        log_with_indent(log, "MIN(%s) = %r", window.expression, min_value)
        log_with_indent(log, "MAX(%s) = %r", window.expression, max_value)

        return min_value, max_value

    def _execute_sql(self, query: str) -> DataFrame:
        return self.spark.sql(query)

    def _target_exist(self, name: str) -> bool:
        from pyspark.sql.functions import col, lower

        log.info("|%s| Checking if table %r exists ...", self.__class__.__name__, name)

        # Do not use SELECT * FROM table, because it may fail if users have no permissions
        # Also we ignore VIEW's as they are not insertable.
        schema, table = name.rsplit(".", maxsplit=1)
        query = f"SHOW TABLES IN {schema} LIKE '{table}'"

        log.debug("|%s| Executing SQL query:", self.__class__.__name__)
        log_lines(log, query, level=logging.DEBUG)

        # Spark normalizes table names to lowercase, so we do the same
        df = self._execute_sql(query).where(lower(col("tableName")) == table.lower())
        if df.take(1):
            log.info("|%s| Table %r exists.", self.__class__.__name__, name)
            return True

        log.info("|%s| Table %r does not exist.", self.__class__.__name__, name)
        return False

    def _save_as_table(
        self,
        df: DataFrame,
        table: str,
        options: IcebergWriteOptions | dict | None = None,
    ) -> None:
        write_options = self.WriteOptions.parse(options)

        writer = df.writeTo(table).using("iceberg")
        for method, value in self._format_write_options(write_options).items():
            if hasattr(writer, method):
                if isinstance(value, Iterable) and not isinstance(value, str):
                    writer = getattr(writer, method)(*value)
                else:
                    writer = getattr(writer, method)(value)
            else:
                writer = writer.option(method, value)

        log.info("|%s| Saving data to a table %r ...", self.__class__.__name__, table)
        writer.createOrReplace()
        log.info("|%s| Table %r is successfully created.", self.__class__.__name__, table)

    def _insert_into(
        self,
        df: DataFrame,
        table: str,
        options: IcebergWriteOptions | dict | None = None,
    ) -> None:
        write_options = self.WriteOptions.parse(options)
        unsupported_options = self._format_write_options(write_options)
        if unsupported_options:
            log.warning(
                "|%s| User-specified options %r are ignored while inserting into existing table. "
                "Using only Iceberg table parameters",
                self.__class__.__name__,
                unsupported_options,
            )

        log.info("|%s| Inserting data into existing table %r ...", self.__class__.__name__, table)

        writer = df.writeTo(table).using("iceberg")

        if write_options.if_exists == IcebergTableExistBehavior.APPEND:
            writer.append()
        else:
            writer.overwritePartitions()

        log.info("|%s| Data is successfully inserted into table %r.", self.__class__.__name__, table)

    def _format_write_options(self, write_options: IcebergWriteOptions) -> dict:
        return write_options.dict(
            by_alias=True,
            exclude_unset=True,
            exclude={"if_exists"},
        )

    def _normalize_table_name(self, table: str) -> str:
        return f"{self.catalog_name}.{table}"



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyspark.sql import SparkSession


def get_hadoop_config(spark_session: SparkSession):
    """
    Get ``org.apache.hadoop.conf.Configuration`` object
    """
    return spark_session.sparkContext._jsc.hadoopConfiguration()  # type: ignore[attr-defined]



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, ClassVar, Optional

from typing_extensions import Literal

try:
    from pydantic.v1 import Field
except (ImportError, AttributeError):
    from pydantic import Field  # type: ignore[no-redef, assignment]

from onetl._util.spark import stringify
from onetl.file.format.file_format import ReadOnlyFileFormat
from onetl.hooks import slot, support_hooks

if TYPE_CHECKING:
    from pyspark.sql import Column, SparkSession
    from pyspark.sql.types import ArrayType, MapType, StructType


PARSE_COLUMN_UNSUPPORTED_OPTIONS = {
    "encoding",
    "lineSep",
    "samplingRatio",
    "primitivesAsString",
    "prefersDecimal",
    "dropFieldIfAllNull",
}


@support_hooks
class JSON(ReadOnlyFileFormat):
    """
    JSON file format. |support_hooks|

    Based on `Spark JSON <https://spark.apache.org/docs/latest/sql-data-sources-json.html>`_ file format.

    Supports reading (but **NOT** writing) files with ``.json`` extension with content like:

    .. code-block:: json
        :caption: example.json

        [
            {"key": "value1"},
            {"key": "value2"}
        ]

    .. versionadded:: 0.9.0

    Examples
    --------

    .. note ::

        You can pass any option mentioned in
        `official documentation <https://spark.apache.org/docs/latest/sql-data-sources-json.html>`_.
        **Option names should be in** ``camelCase``!

        The set of supported options depends on Spark version.

    Reading files:

    .. code:: python

        from onetl.file.format import JSON

        json = JSON(encoding="UTF-8")

    Writing files:

    .. warning::

        Not supported. Use :obj:`JSONLine <onetl.file.format.jsonline.JSONLine>`.

    """

    name: ClassVar[str] = "json"

    multiLine: Literal[True] = True

    encoding: Optional[str] = None
    """
    Encoding of the JSON file.
    Default ``UTF-8``.

    .. note::

        Used only for reading and writing files.

        Ignored by :obj:`~parse_column` and :obj:`~serialize_column` methods.
    """

    lineSep: Optional[str] = None
    """
    Character used to separate lines in the JSON file.

    Defaults:
      * Try to detect for reading (``\\r\\n``, ``\\r``, ``\\n``)
      * ``\\n`` for writing

    .. note::

        Used only for reading and writing files.

        Ignored by :obj:`~parse_column` and :obj:`~serialize_column` methods,
        as they handle each DataFrame row separately.
    """

    allowComments: Optional[bool] = None
    """
    If ``True``, add support for C/C++/Java style comments (``//``, ``/* */``).
    Default ``False``, meaning that JSON files should not contain comments.

    .. note::

        Used only for reading files and :obj:`~parse_column` method.
    """

    allowUnquotedFieldNames: Optional[bool] = None
    """
    If ``True``, allow JSON object field names without quotes (JavaScript-style).
    Default ``False``.

    .. note::

        Used only for reading files and :obj:`~parse_column` method.
    """

    allowSingleQuotes: Optional[bool] = None
    """
    If ``True``, allow JSON object field names to be wrapped with single quotes (``'``).
    Default ``True``.

    .. note::

        Used only for reading files and :obj:`~parse_column` method.
    """

    allowNumericLeadingZeros: Optional[bool] = None
    """
    If ``True``, allow leading zeros in numbers (e.g. ``00012``).
    Default ``False``.

    .. note::

        Used only for reading files and :obj:`~parse_column` method.
    """

    allowNonNumericNumbers: Optional[bool] = None
    """
    If ``True``, allow numbers to contain non-numeric characters, like:
      * scientific notation (e.g. ``12e10``).
      * positive infinity floating point value (``Infinity``, ``+Infinity``, ``+INF``).
      * negative infinity floating point value (``-Infinity``, ``-INF``).
      * Not-a-Number floating point value (``NaN``).

    Default ``True``.

    .. note::

        Used only for reading files and :obj:`~parse_column` method.
    """

    allowBackslashEscapingAnyCharacter: Optional[bool] = None
    """
    If ``True``, prefix ``\\`` can escape any character.
    Default ``False``.

    .. note::

        Used only for reading files and :obj:`~parse_column` method.
    """

    allowUnquotedControlChars: Optional[bool] = None
    """
    If ``True``, allow unquoted control characters (ASCII values 0-31) in strings without escaping them with ``\\``.
    Default ``False``.

    .. note::

        Used only for reading files and :obj:`~parse_column` method.
    """

    mode: Optional[Literal["PERMISSIVE", "DROPMALFORMED", "FAILFAST"]] = None
    """
    How to handle parsing errors:
      * ``PERMISSIVE`` - set field value as ``null``, move raw data to :obj:`~columnNameOfCorruptRecord` column.
      * ``DROPMALFORMED`` - skip the malformed row.
      * ``FAILFAST`` - throw an error immediately.

    Default is ``PERMISSIVE``.

    .. note::

        Used only for reading files and :obj:`~parse_column` method.
    """

    columnNameOfCorruptRecord: Optional[str] = Field(default=None, min_length=1)
    """
    Name of column to put corrupt records in.
    Default is ``_corrupt_record``.

    .. warning::

        If DataFrame schema is provided, this column should be added to schema explicitly:

        .. code:: python

            from onetl.connection import SparkLocalFS
            from onetl.file import FileDFReader
            from onetl.file.format import JSON

            from pyspark.sql.types import StructType, StructField, TimestampType, StringType

            spark = ...

            schema = StructType(
                [
                    StructField("my_field", TimestampType()),
                    StructField("_corrupt_record", StringType()),  # <-- important
                ]
            )

            json = JSON(mode="PERMISSIVE", columnNameOfCorruptRecord="_corrupt_record")

            reader = FileDFReader(
                connection=connection,
                format=json,
                df_schema=schema,  # < ---
            )
            df = reader.run(["/some/file.json"])

    .. note::

        Used only for reading files and :obj:`~parse_column` method.
    """

    samplingRatio: Optional[float] = Field(default=None, ge=0, le=1)
    """
    While inferring schema, read the specified fraction of file rows.
    Default ``1``.

    .. note::

        Used only for reading files. Ignored by :obj:`~parse_column` function.
    """

    primitivesAsString: Optional[bool] = None
    """
    If ``True``, infer all primitive types (string, integer, float, boolean) as strings.
    Default ``False``.

    .. note::

        Used only for reading files. Ignored by :obj:`~parse_column` method.
    """

    prefersDecimal: Optional[bool] = None
    """
    If ``True``, infer all floating-point values as ``Decimal``.
    Default ``False``.

    .. note::

        Used only for reading files. Ignored by :obj:`~parse_column` method.
    """

    dropFieldIfAllNull: Optional[bool] = None
    """
    If ``True`` and inferred column is always null or empty array, exclude if from DataFrame schema.
    Default ``False``.

    .. note::

        Used only for reading files. Ignored by :obj:`~parse_column` method.
    """

    dateFormat: Optional[str] = Field(default=None, min_length=1)
    """
    String format for ``DateType()`` representation.
    Default is ``yyyy-MM-dd``.
    """

    timestampFormat: Optional[str] = Field(default=None, min_length=1)
    """
    String format for `TimestampType()`` representation.
    Default is ``yyyy-MM-dd'T'HH:mm:ss[.SSS][XXX]``.
    """

    timestampNTZFormat: Optional[str] = Field(default=None, min_length=1)
    """
    String format for `TimestampNTZType()`` representation.
    Default is ``yyyy-MM-dd'T'HH:mm:ss[.SSS]``.

    .. note::

        Added in Spark 3.2.0
    """

    timezone: Optional[str] = Field(default=None, min_length=1, alias="timeZone")
    """
    Allows to override timezone used for parsing or serializing date and timestamp values.
    By default, ``spark.sql.session.timeZone`` is used.
    """

    locale: Optional[str] = Field(default=None, min_length=1)
    """
    Locale name used to parse dates and timestamps.
    Default is ``en-US``.

    ..  note::

        Used only for reading files and :obj:`~parse_column` method.
    """

    class Config:
        known_options = frozenset()
        extra = "allow"

    @slot
    def check_if_supported(self, spark: SparkSession) -> None:
        # always available
        pass

    def parse_column(self, column: str | Column, schema: StructType | ArrayType | MapType) -> Column:
        """
        Parses a JSON string column to a structured Spark SQL column using Spark's `from_json <https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.functions.from_json.html>`_ function, based on the provided schema.

        .. versionadded:: 0.11.0

        Parameters
        ----------
        column : str | Column
            The name of the column or the column object containing JSON strings/bytes to parse.

        schema : StructType | ArrayType | MapType
            The schema to apply when parsing the JSON data. This defines the structure of the output DataFrame column.

        Returns
        -------
        Column with deserialized data, with the same structure as the provided schema. Column name is the same as input column.

        Examples
        --------

        >>> from pyspark.sql.types import StructType, StructField, IntegerType, StringType
        >>> from pyspark.sql.functions import decode
        >>> from onetl.file.format import JSON
        >>> df.show()
        +----+--------------------+----------+---------+------+-----------------------+-------------+
        |key |value               |topic     |partition|offset|timestamp              |timestampType|
        +----+--------------------+----------+---------+------+-----------------------+-------------+
        |[31]|[7B 22 6E 61 6D 6...|topicJSON |0        |0     |2024-04-24 16:51:11.739|0            |
        |[32]|[7B 22 6E 61 6D 6...|topicJSON |0        |1     |2024-04-24 16:51:11.749|0            |
        +----+--------------------+----------+---------+------+-----------------------+-------------+
        >>> df.printSchema()
        root
        |-- key: binary (nullable = true)
        |-- value: binary (nullable = true)
        |-- topic: string (nullable = true)
        |-- partition: integer (nullable = true)
        |-- offset: integer (nullable = true)
        |-- timestamp: timestamp (nullable = true)
        |-- timestampType: integer (nullable = true)
        >>> json = JSON()
        >>> json_schema = StructType(
        ...     [
        ...         StructField("name", StringType(), nullable=True),
        ...         StructField("age", IntegerType(), nullable=True),
        ...     ],
        ... )
        >>> parsed_df = df.select(decode("key", "UTF-8").alias("key"), json.parse_column("value", json_schema))
        >>> parsed_df.show()
        +---+-----------+
        |key|value      |
        +---+-----------+
        |1  |{Alice, 20}|
        |2  |  {Bob, 25}|
        +---+-----------+
        >>> parsed_df.printSchema()
        root
        |-- key: string (nullable = true)
        |-- value: struct (nullable = true)
        |    |-- name: string (nullable = true)
        |    |-- age: integer (nullable = true)
        """
        from pyspark.sql import Column, SparkSession  # noqa:  WPS442
        from pyspark.sql.functions import col, from_json

        self.check_if_supported(SparkSession._instantiatedSession)  # noqa:  WPS437
        self._check_unsupported_serialization_options()

        if isinstance(column, Column):
            column_name, column = column._jc.toString(), column.cast("string")  # noqa:  WPS437
        else:
            column_name, column = column, col(column).cast("string")

        options = stringify(self.dict(by_alias=True, exclude_none=True))
        return from_json(column, schema, options).alias(column_name)

    def serialize_column(self, column: str | Column) -> Column:
        """
        Serializes a structured Spark SQL column into a JSON string column using Spark's
        `to_json <https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.functions.to_json.html>`_ function.

        .. versionadded:: 0.11.0

        Parameters
        ----------
        column : str | Column
            The name of the column or the column object containing the data to serialize to JSON format.

        Returns
        -------
        Column with string JSON data. Column name is the same as input column.

        Examples
        --------

        >>> from pyspark.sql.functions import decode
        >>> from onetl.file.format import JSON
        >>> df.show()
        +---+-----------+
        |key|value      |
        +---+-----------+
        |1  |{Alice, 20}|
        |2  |  {Bob, 25}|
        +---+-----------+
        >>> df.printSchema()
        root
        |-- key: string (nullable = true)
        |-- value: struct (nullable = true)
        |    |-- name: string (nullable = true)
        |    |-- age: integer (nullable = true)
        >>> # serializing data into JSON format
        >>> json = JSON()
        >>> serialized_df = df.select("key", json.serialize_column("value"))
        >>> serialized_df.show(truncate=False)
        +---+-------------------------+
        |key|value                    |
        +---+-------------------------+
        |  1|{"name":"Alice","age":20}|
        |  2|{"name":"Bob","age":25}  |
        +---+-------------------------+
        >>> serialized_df.printSchema()
        root
        |-- key: string (nullable = true)
        |-- value: string (nullable = true)
        """
        from pyspark.sql import Column, SparkSession  # noqa:  WPS442
        from pyspark.sql.functions import col, to_json

        self.check_if_supported(SparkSession._instantiatedSession)  # noqa:  WPS437
        self._check_unsupported_serialization_options()

        if isinstance(column, Column):
            column_name = column._jc.toString()  # noqa:  WPS437
        else:
            column_name, column = column, col(column)

        options = stringify(self.dict(by_alias=True, exclude_none=True))
        return to_json(column, options).alias(column_name)

    def _check_unsupported_serialization_options(self):
        current_options = self.dict(by_alias=True, exclude_none=True)
        unsupported_options = current_options.keys() & PARSE_COLUMN_UNSUPPORTED_OPTIONS
        if unsupported_options:
            warnings.warn(
                f"Options `{sorted(unsupported_options)}` are set but not supported "
                f"in `JSON.parse_column` or `JSON.serialize_column`.",
                UserWarning,
                stacklevel=2,
            )

    def __repr__(self):
        options_dict = self.dict(by_alias=True, exclude_none=True, exclude={"multiLine"})
        options_dict = dict(sorted(options_dict.items()))
        options_kwargs = ", ".join(f"{k}={v!r}" for k, v in options_dict.items())
        return f"{self.__class__.__name__}({options_kwargs})"



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from etl_entities.hwm import HWM, ColumnDateHWM, ColumnDateTimeHWM, ColumnIntHWM

if TYPE_CHECKING:
    from pyspark.sql.types import DataType


class SparkTypeToHWM:
    """Registry class for HWM types

    Examples
    --------

    >>> from etl_entities.hwm import ColumnIntHWM, ColumnDateHWM
    >>> from pyspark.sql.types import IntegerType, ShortType, DateType, StringType
    >>> from onetl.hwm.store import SparkTypeToHWM
    >>> SparkTypeToHWM.get(IntegerType())
    <class 'etl_entities.hwm.column.int_hwm.ColumnIntHWM'>
    >>> # multiple type names are supported
    >>> SparkTypeToHWM.get(ShortType())
    <class 'etl_entities.hwm.column.int_hwm.ColumnIntHWM'>
    >>> SparkTypeToHWM.get(DateType())
    <class 'etl_entities.hwm.column.date_hwm.ColumnDateHWM'>
    >>> SparkTypeToHWM.get(StringType())
    """

    _mapping: ClassVar[dict[DataType | type[DataType], type[HWM]]] = {}

    @classmethod
    def get(cls, spark_type: DataType) -> type[HWM] | None:
        # avoid importing pyspark in the module
        from pyspark.sql.types import (  # noqa: WPS235
            ByteType,
            DateType,
            DecimalType,
            DoubleType,
            FloatType,
            FractionalType,
            IntegerType,
            LongType,
            NumericType,
            ShortType,
            TimestampType,
        )

        default_mapping: dict[type[DataType], type[HWM]] = {
            ByteType: ColumnIntHWM,
            IntegerType: ColumnIntHWM,
            ShortType: ColumnIntHWM,
            LongType: ColumnIntHWM,
            DateType: ColumnDateHWM,
            TimestampType: ColumnDateTimeHWM,
            # for Oracle which does not differ between int and float/double - everything is Decimal
            FloatType: ColumnIntHWM,
            DoubleType: ColumnIntHWM,
            DecimalType: ColumnIntHWM,
            FractionalType: ColumnIntHWM,
            NumericType: ColumnIntHWM,
        }

        return (
            cls._mapping.get(spark_type)
            or cls._mapping.get(spark_type.__class__)
            or default_mapping.get(spark_type.__class__)
        )

    @classmethod
    def add(cls, spark_type: DataType | type[DataType], klass: type[HWM]) -> None:
        cls._mapping[spark_type] = klass


def register_spark_type_to_hwm_type_mapping(*spark_types: DataType | type[DataType]):
    """Decorator for registering mapping between Spark data type and HWM type.

    Accepts both data type class and instance.

    Examples
    --------

    >>> from etl_entities.hwm import ColumnHWM
    >>> from onetl.hwm.store import SparkTypeToHWM
    >>> from onetl.hwm.store import SparkTypeToHWM, register_spark_type_to_hwm_type_mapping
    >>> from pyspark.sql.types import IntegerType, DecimalType
    >>> @register_spark_type_to_hwm_type_mapping(IntegerType, DecimalType(38, 0))
    ... class MyHWM(ColumnHWM): ...
    >>> SparkTypeToHWM.get(IntegerType())
    <class 'onetl.hwm.store.hwm_class_registry.MyHWM'>
    >>> SparkTypeToHWM.get(DecimalType(38, 0))
    <class 'onetl.hwm.store.hwm_class_registry.MyHWM'>
    >>> SparkTypeToHWM.get(DecimalType(38, 10))
    <class 'etl_entities.hwm.column.int_hwm.ColumnIntHWM'>
    """

    def wrapper(cls: type[HWM]):
        for spark_type in spark_types:
            SparkTypeToHWM.add(spark_type, cls)
        return cls

    return wrapper



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, ClassVar, Optional

try:
    from pydantic.v1 import ByteSize, SecretStr
except (ImportError, AttributeError):
    from pydantic import ByteSize, SecretStr  # type: ignore[no-redef, assignment]

from onetl._util.java import try_import_java_class
from onetl._util.scala import get_default_scala_version
from onetl._util.spark import get_spark_version
from onetl._util.version import Version
from onetl.exception import MISSING_JVM_CLASS_MSG
from onetl.file.format.file_format import ReadWriteFileFormat
from onetl.hooks import slot, support_hooks

if TYPE_CHECKING:
    from pyspark.sql import DataFrameReader, SparkSession

log = logging.getLogger(__name__)


@support_hooks
class Excel(ReadWriteFileFormat):
    """
    Excel file format. |support_hooks|

    Based on `Spark Excel <https://github.com/crealytics/spark-excel>`_ file format.

    Supports reading/writing files with ``.xlsx`` (read/write) and ``.xls`` (read only) extensions.

    .. dropdown:: Version compatibility

        * Spark versions: 3.2.x - 4.0.x

            .. warning::

                Not all combinations of Spark version and package version are supported.
                See `Maven index <https://mvnrepository.com/artifact/dev.mauch/spark-excel>`_
                and `official documentation <https://github.com/crealytics/spark-excel>`_.

        * Java versions: 8 - 22

        See documentation from link above.

    .. versionadded:: 0.9.4

    Examples
    --------

    .. note ::

        You can pass any option mentioned in
        `official documentation <https://github.com/crealytics/spark-excel>`_.
        **Option names should be in** ``camelCase``!

        The set of supported options depends on ``spark-excel`` package version.

    .. tabs::

        .. code-tab:: py Reading files

            from pyspark.sql import SparkSession
            from onetl.file.format import Excel

            # Create Spark session with Excel package loaded
            maven_packages = Excel.get_packages(
                package_version="0.31.2",
                spark_version="3.5.6",
            )
            spark = (
                SparkSession.builder.appName("spark-app-name")
                .config("spark.jars.packages", ",".join(maven_packages))
                .getOrCreate()
            )

            excel = Excel(header=True, inferSchema=True)

        .. code-tab:: py Writing files

            # Create Spark session with Excel package loaded
            spark = ...

            from onetl.file.format import XML

            excel = Excel(header=True, dataAddress="'Sheet1'!A1")
    """

    name: ClassVar[str] = "excel"

    header: bool = False
    """
    If ``True``, the first row in file is conditioned as a header.
    Default ``False``.
    """

    dataAddress: Optional[str] = None
    """
    Cell address used as starting point.
    For example: ``'A1'`` or ``'Sheet1'!A1``
    """

    timestampFormat: Optional[str] = None
    """
    Format string used for parsing or serializing timestamp values.
    Default ``yyyy-mm-dd hh:mm:ss[.fffffffff]``.
    """

    dateFormat: Optional[str] = None
    """
    Format string used for parsing or serializing date values.
    Default ``yyyy-MM-dd``.
    """

    treatEmptyValuesAsNulls: Optional[bool] = None
    """
    If ``True``, empty cells are parsed as ``null`` values.
    If ``False``, empty cells are parsed as empty strings.
    Default ``True``.

    .. note::

        Used only for reading files.
    """

    setErrorCellsToFallbackValues: Optional[bool] = None
    """
    If ``True``, cells containing ``#N/A`` value are replaced with default value for column type,
    e.g. 0 for ``IntegerType()``. If ``False``, ``#N/A`` values are replaced with ``null``.
    Default ``False``.

    .. note::

        Used only for reading files.
    """

    usePlainNumberFormat: Optional[bool] = None
    """
    If ``True``, read or write numeric values with plain format, without using scientific notation or rounding.
    Default ``False``.
    """

    inferSchema: Optional[bool] = None
    """
    If ``True``, infer DataFrame schema based on cell content.
    If ``False`` and no explicit DataFrame schema is passed, all columns are ``StringType()``.

    .. note::

        Used only for reading files.
    """

    workbookPassword: Optional[SecretStr] = None
    """
    If Excel file is encrypted, provide password to open it.

    .. note::

        Used only for reading files. Cannot be used to write files.
    """

    maxRowsInMemory: Optional[int] = None
    """
    If set, use streaming reader and fetch only specified number of rows per iteration.
    This reduces memory usage for large files.
    Default ``None``, which means reading the entire file content to memory.

    .. warning::

        Can be used only with ``.xlsx`` files, but fails on ``.xls``.

    .. note::

        Used only for reading files.
    """

    maxByteArraySize: Optional[ByteSize] = None
    """
    If set, overrides memory limit (in bytes) of byte array size used for reading rows from input file.
    Default ``0``, which means using default limit.

    See `IOUtils.setByteArrayMaxOverride <https://poi.apache.org/apidocs/5.0/org/apache/poi/util/IOUtils.html#setByteArrayMaxOverride-int->`_
    documentation.

    .. note::

        Used only for reading files.
    """

    tempFileThreshold: Optional[ByteSize] = None
    """
    If value is greater than 0, large zip entries will be written to temporary files after reaching this threshold.
    If value is 0, all zip entries will be written to temporary files.
    If value is -1, no temp files will be created, which may cause errors if zip entry is larger than 2GiB.

    .. note::

        Used only for reading files.
    """

    excerptSize: Optional[int] = None
    """
    If ``inferSchema=True``, set number of rows to infer schema from.
    Default ``10``.

    .. note::

        Used only for reading files.
    """

    class Config:
        known_options = frozenset()
        extra = "allow"

    @slot
    @classmethod
    def get_packages(
        cls,
        package_version: str,
        spark_version: str,
        scala_version: str | None = None,
    ) -> list[str]:
        """
        Get package names to be downloaded by Spark. |support_hooks|

        .. warning::

            Not all combinations of Spark version and package version are supported.
            See `Maven index <https://mvnrepository.com/artifact/dev.mauch/spark-excel>`_
            and `official documentation <https://github.com/crealytics/spark-excel>`_.

        .. versionadded:: 0.9.4
        .. versionchanged:: 0.14.0
            Maven package ``com.crealytics:spark-excel`` was renamed to ``dev.mauch:spark-excel``.

        Parameters
        ----------
        package_version : str
            Package version in format ``major.minor.patch``.

            .. versionchanged:: 0.14.0
                This parameter is now mandatory.

        spark_version : str
            Spark version in format ``major.minor.patch``.

        scala_version : str, optional
            Scala version in format ``major.minor``.

            If ``None``, ``spark_version`` is used to determine Scala version.

        Examples
        --------

        .. code:: python

            from onetl.file.format import Excel

            Excel.get_packages(
                package_version="0.31.2",
                spark_version="3.5.6",
            )
            Excel.get_packages(
                package_version="0.31.2",
                spark_version="3.5.6",
                scala_version="2.12",
            )

        """

        version = Version(package_version)
        if version < Version("0.30"):
            raise ValueError(f"Package version should be at least 0.30, got {package_version}")

        spark_ver = Version(spark_version).min_digits(3)
        scala_ver = Version(scala_version).min_digits(2) if scala_version else get_default_scala_version(spark_ver)
        return [
            f"dev.mauch:spark-excel_{scala_ver.format('{0}.{1}')}:{spark_ver.format('{0}.{1}.{2}')}_{version}",
        ]

    @slot
    def check_if_supported(self, spark: SparkSession) -> None:
        java_class = "dev.mauch.spark.excel.v2.ExcelDataSource"

        try:
            try_import_java_class(spark, java_class)
        except Exception as e:
            spark_version = get_spark_version(spark).format("{0}.{1}.{2}")
            msg = MISSING_JVM_CLASS_MSG.format(
                java_class=java_class,
                package_source=self.__class__.__name__,
                args=f"spark_version='{spark_version}'",
            )
            raise ValueError(msg) from e

    @slot
    def apply_to_reader(self, reader: DataFrameReader) -> DataFrameReader:
        options = self.dict(by_alias=True, exclude_none=True)
        if self.workbookPassword:
            options["workbookPassword"] = self.workbookPassword.get_secret_value()
        return reader.format(self.name).options(**options)

    def __repr__(self):
        options_dict = self.dict(by_alias=True, exclude_none=True)
        options_dict = dict(sorted(options_dict.items()))
        options_kwargs = ", ".join(f"{k}={v!r}" for k, v in options_dict.items())
        return f"{self.__class__.__name__}({options_kwargs})"



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import logging
import warnings
from typing import TYPE_CHECKING, ClassVar, Optional, Union

from typing_extensions import Literal

try:
    from pydantic.v1 import Field
except (ImportError, AttributeError):
    from pydantic import Field  # type: ignore[no-redef, assignment]

from onetl._util.java import try_import_java_class
from onetl._util.scala import get_default_scala_version
from onetl._util.spark import get_spark_version, stringify
from onetl._util.version import Version
from onetl.exception import MISSING_JVM_CLASS_MSG
from onetl.file.format.file_format import ReadWriteFileFormat
from onetl.hooks import slot, support_hooks

if TYPE_CHECKING:
    from pyspark.sql import Column, SparkSession
    from pyspark.sql.types import StructType


log = logging.getLogger(__name__)
PARSE_COLUMN_UNSUPPORTED_OPTIONS = {"inferSchema", "samplingRatio"}


@support_hooks
class XML(ReadWriteFileFormat):
    """
    XML file format. |support_hooks|

    Based on `Databricks Spark XML <https://github.com/databricks/spark-xml>`_ file format.

    Supports reading/writing files with ``.xml`` extension.

    .. versionadded:: 0.9.5

    .. dropdown:: Version compatibility

        * Spark versions: 3.2.x - 3.5.x
        * Java versions: 8 - 20

        See `official documentation <https://github.com/databricks/spark-xml>`_.

    Examples
    --------

    .. note ::

        You can pass any option mentioned in
        `official documentation <https://github.com/databricks/spark-xml>`_.
        **Option names should be in** ``camelCase``!

        The set of supported options depends on ``spark-xml`` version.

    .. tabs::

        .. code-tab:: py Reading files

            from onetl.file.format import XML
            from pyspark.sql import SparkSession

            # Create Spark session with XML package loaded
            maven_packages = XML.get_packages(spark_version="3.5.6")
            spark = (
                SparkSession.builder.appName("spark-app-name")
                .config("spark.jars.packages", ",".join(maven_packages))
                .getOrCreate()
            )

            xml = XML(rowTag="item", mode="PERMISSIVE")

        .. tab:: Writing files

            .. warning::

                Due to `bug <https://github.com/databricks/spark-xml/issues/664>`_ written files currently does not have ``.xml`` extension.

            .. code:: python

                # Create Spark session with XML package loaded
                spark = ...

                from onetl.file.format import XML

                xml = XML(rowTag="item", rootTag="data", compression="gzip")

    """

    name: ClassVar[str] = "xml"

    row_tag: str = Field(alias="rowTag")
    """
    XML tag that encloses each row in XML. Required.
    """

    rootTag: Optional[str] = None
    """
    XML tag that encloses content of all DataFrame. Default is ``ROWS``.

    .. note::

        Used only for writing files.
    """

    compression: Union[str, Literal["bzip2", "gzip", "lz4", "snappy"], None] = None
    """
    Compression codec. By default no compression is used.

    .. note::

        Used only for writing files.
    """

    timestampFormat: Optional[str] = None
    """
    Format string used for parsing or serializing timestamp values.
    By default, ISO 8601 format is used (``yyyy-MM-ddTHH:mm:ss.SSSZ``).
    """

    dateFormat: Optional[str] = None
    """
    Format string used for parsing or serializing date values.
    By default, ISO 8601 format is used (``yyyy-MM-dd``).
    """

    timezone: Optional[str] = None
    """
    Allows to override timezone used for parsing or serializing date and timestamp values.
    By default, ``spark.sql.session.timeZone`` is used.
    """

    nullValue: Optional[str] = None
    """
    String value used to represent null. Default is ``null`` string.
    """

    ignoreSurroundingSpaces: Optional[bool] = None
    """
    If ``True``, trim surrounding spaces while parsing values. Default ``false``.

    .. note::

        Used only for reading files or by :obj:`~parse_column` function.
    """

    mode: Optional[Literal["PERMISSIVE", "DROPMALFORMED", "FAILFAST"]] = None
    """
    How to handle parsing errors:
      * ``PERMISSIVE`` - set field value as ``null``, move raw data to :obj:`~columnNameOfCorruptRecord` column.
      * ``DROPMALFORMED`` - skip the malformed row.
      * ``FAILFAST`` - throw an error immediately.

    Default is ``PERMISSIVE``.

    .. note::

        Used only for reading files or by :obj:`~parse_column` function.
    """

    columnNameOfCorruptRecord: Optional[str] = None
    """
    Name of DataFrame column there corrupted row is stored with ``mode=PERMISSIVE``.

    .. warning::

        If DataFrame schema is provided, this column should be added to schema explicitly:

        .. code:: python

            from onetl.connection import SparkLocalFS
            from onetl.file import FileDFReader
            from onetl.file.format import XML

            from pyspark.sql.types import StructType, StructField, TImestampType, StringType

            spark = ...
            schema = StructType(
                [
                    StructField("my_field", TimestampType()),
                    StructField("_corrupt_record", StringType()),  # <-- important
                ]
            )
            xml = XML(rowTag="item", columnNameOfCorruptRecord="_corrupt_record")

            reader = FileDFReader(
                connection=connection,
                format=xml,
                df_schema=schema,  # < ---
            )
            df = reader.run(["/some/file.xml"])

    .. note::

        Used only for reading files or by :obj:`~parse_column` function.
    """

    inferSchema: Optional[bool] = None
    """
    If ``True``, try to infer the input schema by reading a sample of the file (see :obj:`~samplingRatio`).
    Default ``False`` which means that all parsed columns will be ``StringType()``.

    .. note::

        Used only for reading files. Ignored by :obj:`~parse_column` function.
    """

    samplingRatio: Optional[float] = Field(default=None, ge=0, le=1)
    """
    For ``inferSchema=True``, read the specified fraction of rows to infer the schema.
    Default ``1``.

    .. note::

        Used only for reading files. Ignored by :obj:`~parse_column` function.
    """

    charset: Optional[str] = None
    """
    File encoding. Default is ``UTF-8``

    .. note::

        Used only for reading files or by :obj:`~parse_column` function.
    """

    valueTag: Optional[str] = None
    """
    Value used to replace missing values while parsing attributes like ``<sometag someattr>``.
    Default ``_VALUE``.

    .. note::

        Used only for reading files or by :obj:`~parse_column` function.
    """

    attributePrefix: Optional[str] = None
    """
    While parsing tags containing attributes like ``<sometag attr="value">``, attributes are stored as
    DataFrame schema columns with specified prefix, e.g. ``_attr``.
    Default ``_``.

    .. note::

        Used only for reading files or by :obj:`~parse_column` function.
    """

    excludeAttribute: Optional[bool] = None
    """
    If ``True``, exclude attributes while parsing tags like ``<sometag attr="value">``.
    Default ``false``.

    .. note::

        Used only for reading files or by :obj:`~parse_column` function.
    """

    wildcardColName: Optional[str] = None
    """
    Name of column or columns which should be preserved as raw XML string, and not parsed.

    .. warning::

        If DataFrame schema is provided, this column should be added to schema explicitly.
        See :obj:`~columnNameOfCorruptRecord` example.

    .. note::

        Used only for reading files or by :obj:`~parse_column` function.
    """

    ignoreNamespace: Optional[bool] = None
    """
    If ``True``, all namespaces like ``<ns:tag>`` will be ignored and treated as just ``<tag>``.
    Default ``False``.

    .. note::

        Used only for reading files or by :obj:`~parse_column` function.
    """

    rowValidationXSDPath: Optional[str] = None
    """
    Path to XSD file which should be used to validate each row.
    If row does not match XSD, it will be treated as error, behavior depends on :obj:`~mode` value.

    Default is no validation.

    .. note::

        If Spark session is created with ``master=yarn`` or ``master=k8s``, XSD
        file should be accessible from all Spark nodes. This can be achieved by calling:

        .. code:: python

            spark.addFile("/path/to/file.xsd")

        And then by passing ``rowValidationXSDPath=file.xsd`` (relative path).

    .. note::

        Used only for reading files or by :obj:`~parse_column` function.
    """

    declaration: Optional[str] = None
    """
    Content of `<?XML ... ?>` declaration.
    Default is ``version="1.0" encoding="UTF-8" standalone="yes"``.

    .. note::

        Used only for writing files.
    """

    arrayElementName: Optional[str] = None
    """
    If DataFrame column is ``ArrayType``, its content will be written to XML
    inside ``<arrayElementName>...</arrayElementName>`` tag.
    Default is ``item``.

    .. note::

        Used only for writing files.
    """

    class Config:
        known_options = frozenset()
        prohibited_options = frozenset(("path",))  # filled by FileDFReader/FileDFWriter
        extra = "allow"

    @slot
    @classmethod
    def get_packages(  # noqa: WPS231
        cls,
        spark_version: str,
        scala_version: str | None = None,
        package_version: str | None = None,
    ) -> list[str]:
        """
        Get package names to be downloaded by Spark. |support_hooks|

        .. versionadded:: 0.9.5

        Parameters
        ----------
        spark_version : str
            Spark version in format ``major.minor.patch``.

        scala_version : str, optional
            Scala version in format ``major.minor``.

            If ``None``, ``spark_version`` is used to determine Scala version.

        package_version : str, optional
            Package version in format ``major.minor.patch``. Default is ``0.18.0``.

            See `Maven index <https://mvnrepository.com/artifact/com.databricks/spark-xml>`_
            for list of available versions.

            .. warning::

                Version ``0.13`` and below are not supported.

            .. note::

                It is not guaranteed that custom package versions are supported.
                Tests are performed only for default version.

        Examples
        --------

        .. code:: python

            from onetl.file.format import XML

            XML.get_packages(spark_version="3.5.6")
            XML.get_packages(spark_version="3.5.6", scala_version="2.12")
            XML.get_packages(
                spark_version="3.5.6",
                scala_version="2.12",
                package_version="0.18.0",
            )

        """
        spark_ver = Version(spark_version)
        if spark_ver.major >= 4:
            # since Spark 4.0, XML is bundled with Spark
            return []

        if package_version:
            version = Version(package_version).min_digits(3)
            if version < Version("0.14"):
                raise ValueError(f"Package version must be above 0.13, got {version}")
            log.warning("Passed custom package version %r, it is not guaranteed to be supported", package_version)
        else:
            version = Version("0.18.0")

        scala_ver = Version(scala_version).min_digits(2) if scala_version else get_default_scala_version(spark_ver)
        return [f"com.databricks:spark-xml_{scala_ver.format('{0}.{1}')}:{version}"]

    @slot
    def check_if_supported(self, spark: SparkSession) -> None:
        version = get_spark_version(spark)
        if version.major >= 4:
            # since Spark 4.0, XML is bundled with Spark
            return

        java_class = "com.databricks.spark.xml.XmlReader"
        try:
            try_import_java_class(spark, java_class)
        except Exception as e:
            spark_version = get_spark_version(spark).format("{0}.{1}")
            msg = MISSING_JVM_CLASS_MSG.format(
                java_class=java_class,
                package_source=self.__class__.__name__,
                args=f"spark_version='{spark_version}'",
            )
            raise ValueError(msg) from e

    def parse_column(self, column: str | Column, schema: StructType) -> Column:
        """
        Parses an XML string column into a structured Spark SQL column using the ``from_xml`` function
        provided by the `Databricks Spark XML library <https://github.com/databricks/spark-xml#pyspark-notes>`_
        based on the provided schema.

        .. note::

            This method assumes that the ``spark-xml`` package is installed: :obj:`~get_packages`.

        .. note::

            This method parses each DataFrame row individually. Therefore, for a specific column, each row must contain exactly one occurrence of the ``rowTag`` specified.
            If your XML data includes a root tag that encapsulates multiple row tags, you can adjust the schema to use an ``ArrayType`` to keep all child elements under the single root.

            .. code-block:: xml

                <books>
                    <book><title>Book One</title><author>Author A</author></book>
                    <book><title>Book Two</title><author>Author B</author></book>
                </books>

            And the corresponding schema in Spark using an ``ArrayType``:

            .. code-block:: python

                from pyspark.sql.types import StructType, StructField, ArrayType, StringType
                from onetl.file.format import XML

                # each DataFrame row has exactly one <books> tag
                xml = XML(rowTag="books")
                # each <books> tag have multiple <book> tags, so using ArrayType for such field
                schema = StructType(
                    [
                        StructField(
                            "book",
                            ArrayType(
                                StructType(
                                    [
                                        StructField("title", StringType(), nullable=True),
                                        StructField("author", StringType(), nullable=True),
                                    ],
                                ),
                            ),
                            nullable=True,
                        ),
                    ],
                )

        .. versionadded:: 0.11.0

        Parameters
        ----------
        column : str | Column
            The name of the column or the column object containing XML strings/bytes to parse.

        schema : StructType
            The schema to apply when parsing the XML data. This defines the structure of the output DataFrame column.

        Returns
        -------
        Column with deserialized data, with the same structure as the provided schema. Column name is the same as input column.

        Examples
        --------

        >>> from pyspark.sql.types import StructType, StructField, IntegerType, StringType
        >>> from onetl.file.format import XML
        >>> df.show()
        +--+------------------------------------------------+
        |id|value                                           |
        +--+------------------------------------------------+
        |1 |<person><name>Alice</name><age>20</age></person>|
        |2 |<person><name>Bob</name><age>25</age></person>  |
        +--+------------------------------------------------+
        >>> df.printSchema()
        root
        |-- id: integer (nullable = true)
        |-- value: string (nullable = true)
        >>> xml = XML(rowTag="person")
        >>> xml_schema = StructType(
        ...     [
        ...         StructField("name", StringType(), nullable=True),
        ...         StructField("age", IntegerType(), nullable=True),
        ...     ],
        ... )
        >>> parsed_df = df.select("key", xml.parse_column("value", xml_schema))
        >>> parsed_df.show()
        +--+-----------+
        |id|value      |
        +--+-----------+
        |1 |{Alice, 20}|
        |2 |  {Bob, 25}|
        +--+-----------+
        >>> parsed_df.printSchema()
        root
        |-- id: integer (nullable = true)
        |-- value: struct (nullable = true)
        |    |-- name: string (nullable = true)
        |    |-- age: integer (nullable = true)
        """
        from pyspark.sql import Column, SparkSession  # noqa: WPS442

        spark = SparkSession._instantiatedSession  # noqa: WPS437
        self.check_if_supported(spark)
        self._check_unsupported_serialization_options()

        from pyspark.sql.functions import col

        if isinstance(column, Column):
            column_name, column = column._jc.toString(), column.cast("string")  # noqa: WPS437
        else:
            column_name, column = column, col(column).cast("string")

        options = self.dict(by_alias=True, exclude_none=True)
        version = get_spark_version(spark)
        if version.major >= 4:
            from pyspark.sql.functions import from_xml  # noqa: WPS450

            return from_xml(column, schema, stringify(options)).alias(column_name)

        from pyspark.sql.column import _to_java_column  # noqa: WPS450

        java_column = _to_java_column(column)
        java_schema = spark._jsparkSession.parseDataType(schema.json())  # noqa: WPS437
        scala_options = spark._jvm.org.apache.spark.api.python.PythonUtils.toScalaMap(  # noqa: WPS219, WPS437
            stringify(options),
        )
        jc = spark._jvm.com.databricks.spark.xml.functions.from_xml(  # noqa: WPS219, WPS437
            java_column,
            java_schema,
            scala_options,
        )
        return Column(jc).alias(column_name)

    def _check_unsupported_serialization_options(self):
        current_options = self.dict(by_alias=True, exclude_none=True)
        unsupported_options = current_options.keys() & PARSE_COLUMN_UNSUPPORTED_OPTIONS
        if unsupported_options:
            warnings.warn(
                f"Options `{sorted(unsupported_options)}` are set but not supported in `XML.parse_column`.",
                UserWarning,
                stacklevel=2,
            )

    def __repr__(self):
        options_dict = self.dict(by_alias=True, exclude_none=True)
        options_dict = dict(sorted(options_dict.items()))
        options_kwargs = ", ".join(f"{k}={v!r}" for k, v in options_dict.items())
        return f"{self.__class__.__name__}({options_kwargs})"



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from typing_extensions import Literal

from onetl.file.format.file_format import ReadWriteFileFormat
from onetl.hooks import slot, support_hooks

if TYPE_CHECKING:
    from pyspark.sql import SparkSession

PROHIBITED_OPTIONS = frozenset(
    (
        # These options should be passed in Spark session config, not file format options
        "spark.*",
    ),
)

ORC_JAVA_OPTIONS = frozenset(
    ("orc.*",),
)


@support_hooks
class ORC(ReadWriteFileFormat):
    """
    ORC file format (columnar). |support_hooks|

    Based on `Spark ORC Files <https://spark.apache.org/docs/latest/sql-data-sources-orc.html>`_ file format.

    Supports reading/writing files with ``.orc`` extension.

    .. versionadded:: 0.9.0

    Examples
    --------

    .. note ::

        You can pass any option mentioned in
        `official documentation <https://spark.apache.org/docs/latest/sql-data-sources-orc.html>`_.
        **Option names should be in** ``camelCase``!

        The set of supported options depends on Spark version.

        You may also set options mentioned `orc-java documentation <https://orc.apache.org/docs/core-java-config.html>`_.
        They are prefixed with ``orc.`` with dots in names, so instead of calling constructor ``ORC(orc.option=True)`` (invalid in Python)
        you should call method ``ORC.parse({"orc.option": True})``.

    .. tabs::

        .. code-tab:: py Reading files

            from onetl.file.format import ORC

            orc = ORC(mergeSchema=True)

        .. tab:: Writing files

            .. code:: python

                from onetl.file.format import ORC

                orc = ORC.parse(
                    {
                        "compression": "snappy",
                        # Enable Bloom filter for columns 'id' and 'name'
                        "orc.bloom.filter.columns": "id,name",
                        # Set Bloom filter false positive probability
                        "orc.bloom.filter.fpp": 0.01,
                        # Do not use dictionary for 'highly_selective_column'
                        "orc.column.encoding.direct": "highly_selective_column",
                        # other options
                    }
                )

    """

    name: ClassVar[str] = "orc"

    mergeSchema: Optional[bool] = None
    """
    Merge schemas of all ORC files being read into a single schema.
    By default, Spark config option ``spark.sql.orc.mergeSchema`` value is used (``False``).

    .. note::

        Used only for reading files.
    """

    compression: Union[
        str,
        Literal["uncompressed", "snappy", "zlib", "lzo", "zstd", "lz4"],
        None,
    ] = None
    """
    Compression codec of the ORC files.
    By default, Spark config option ``spark.sql.orc.compression.codec`` value is used (``snappy``).

    .. note::

        Used only for writing files.
    """

    class Config:
        known_options = ORC_JAVA_OPTIONS
        prohibited_options = PROHIBITED_OPTIONS
        extra = "allow"

    @slot
    def check_if_supported(self, spark: SparkSession) -> None:
        # always available
        pass

    def __repr__(self):
        options_dict = self.dict(by_alias=True, exclude_none=True)
        options_dict = dict(sorted(options_dict.items()))
        if any("." in field for field in options_dict.keys()):
            return f"{self.__class__.__name__}.parse({options_dict})"

        options_kwargs = ", ".join(f"{k}={v!r}" for k, v in options_dict.items())
        return f"{self.__class__.__name__}({options_kwargs})"



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from typing_extensions import Literal

from onetl.file.format.file_format import ReadWriteFileFormat
from onetl.hooks import slot, support_hooks

if TYPE_CHECKING:
    from pyspark.sql import SparkSession

PROHIBITED_OPTIONS = frozenset(
    (
        # These options should be passed in Spark session config, not file format options
        "spark.*",
    ),
)

PARQUET_LIBRARY_OPTIONS = frozenset(
    ("parquet.*",),
)


@support_hooks
class Parquet(ReadWriteFileFormat):
    """
    Parquet file format (columnar). |support_hooks|

    Based on `Spark Parquet Files <https://spark.apache.org/docs/latest/sql-data-sources-parquet.html>`_ file format.

    Supports reading/writing files with ``.parquet`` extension.

    .. versionadded:: 0.9.0

    Examples
    --------

    .. note ::

        You can pass any option mentioned in
        `official documentation <https://spark.apache.org/docs/latest/sql-data-sources-parquet.html>`_.
        **Option names should be in** ``camelCase``!

        The set of supported options depends on Spark version.

        You may also set options mentioned `parquet-hadoop documentation <https://github.com/apache/parquet-java/blob/master/parquet-hadoop/README.md>`_.
        They are prefixed with ``parquet.`` with dots in names, so instead of calling constructor ``Parquet(parquet.option=True)`` (invalid in Python)
        you should call method ``Parquet.parse({"parquet.option": True})``.

    .. tabs::

        .. code-tab:: py Reading files

            from onetl.file.format import Parquet

            parquet = Parquet(mergeSchema=True)

        .. code-tab:: py Writing files

            from onetl.file.format import Parquet

            parquet = Parquet.parse(
                {
                    "compression": "snappy",
                    # Enable Bloom filter for columns 'id' and 'name'
                    "parquet.bloom.filter.enabled#id": True,
                    "parquet.bloom.filter.enabled#name": True,
                    # Set expected number of distinct values for column 'id'
                    "parquet.bloom.filter.expected.ndv#id": 10_000_000,
                    # other options
                }
            )

    """

    name: ClassVar[str] = "parquet"

    mergeSchema: Optional[bool] = None
    """
    Merge schemas of all Parquet files being read into a single schema.
    By default, Spark config option ``spark.sql.parquet.mergeSchema`` value is used (``false``).

    .. note::

        Used only for reading files.
    """

    compression: Union[
        str,
        Literal["uncompressed", "snappy", "gzip", "lzo", "brotli", "lz4", "lz4raw", "zstd"],
        None,
    ] = None
    """
    Compression codec of the Parquet files.
    By default, Spark config option ``spark.sql.parquet.compression.codec`` value is used (``snappy``).

    .. note::

        Used only for writing files.
    """

    class Config:
        known_options = PARQUET_LIBRARY_OPTIONS
        prohibited_options = PROHIBITED_OPTIONS
        extra = "allow"

    @slot
    def check_if_supported(self, spark: SparkSession) -> None:
        # always available
        pass

    def __repr__(self):
        options_dict = self.dict(by_alias=True, exclude_none=True)
        options_dict = dict(sorted(options_dict.items()))
        if any("." in field for field in options_dict.keys()):
            return f"{self.__class__.__name__}.parse({options_dict})"

        options_kwargs = ", ".join(f"{k}={v!r}" for k, v in options_dict.items())
        return f"{self.__class__.__name__}({options_kwargs})"



# SPDX-FileCopyrightText: 2021-2024 MTS PJSC
# SPDX-License-Identifier: Apache-2.0
from onetl.impl import GenericOptions


class IcebergExtra(GenericOptions):
    """
    This class is responsible for validating additional options that are passed from the user
    to the Iceberg connection. These extra options are configurations that can be provided to the
    Iceberg, which aren't part of the core connection options.

    See `documentation <https://iceberg.apache.org/docs/latest/spark-configuration/>`_
    for more details
    """

    class Config:
        extra = "allow"
